"""
File Purpose: tools for saving/reading xarrays.
(xarray builtins fail for coords with dtype=object.)
"""
import ast
import os
import re

import pandas as pd
import xarray as xr

from .xarray_accessors import pcAccessor
from ..history import code_snapshot_info
from ...errors import (
    InputError, InputConflictError,
    SerializingError, DeserializingError,
)


''' --------------------- Serializing --------------------- '''

class XarrayIoSerializable():
    '''object that can be serialized when saving/reading xarrays.
    obj == XarrayIoSerializable.deserialize(obj.serialize()).
    obj.serialize() creates a dict d where ast.literal_eval(str(d)) == d,
        and also d['typename'] == obj._serial_typename,
            so that deserialize knows which type of object to create.
    '''
    SERIAL_TYPES = {}  # dict of {serialization name: cls}

    # # # SUBCLASS SHOULD IMPLEMENT # # #
    def to_dict(self):
        '''return dictionary of info about self.
        subclass must override this or override self.serialize to avoid using self.to_dict
        '''
        raise NotImplementedError(f"{type(self).__name__}.to_dict")

    # # # SUBCLASS MAY WISH TO OVERRIDE # # #
    @classmethod
    def from_dict(cls, d):
        '''return cls from dict d. 
        If 'd' contains 'typename', result is like cls.deserialize(d) instead,
            however, first require that result is an instance of cls (or subclass of cls).
            E.g. Fluid.from_dict(dict(typename='EppicDist', ...)) returns EppicDist,
            but Snap.from_dict(dict(typename='EppicDist', ...)) crashes with InputError.

        Subclasses wishing to alter deserialize behavior will probably override this method instead.
        '''
        if 'typename' in d:
            result = cls.deserialize(d)
            if isinstance(result, cls):
                return result
            else:
                errmsg = (f'cls.from_dict(d), with "typename" in d expected typename corresponding to subclass of cls,\n'
                          f'but got typename={d["typename"]!r}, which corresponds to type {type(result).__name__}, '
                          f'which is not a subclass of cls={cls.__name__}.\n'
                          f'To produce arbitrary XarrayIoSerializable objects, deserialize() instead of from_dict().')
                raise InputError(errmsg)
        else:
            return cls(**d)

    # # # SERIALIZATION BEHAVIOR DEFINED HERE # # #
    def __init_subclass__(cls, *, _serial_typename=None, **kw):
        '''register cls as a subclass of XarrayIoSerializable.

        _serial_typename: None or str
            name to use for this type, for serialization purposes.
            None --> use cls.__name__.
            This method also sets cls._serial_typename = _serial_typename.
        '''
        super().__init_subclass__(**kw)
        XarrayIoSerializable.register_type(cls, _serial_typename=_serial_typename)

    @staticmethod
    def register_type(cls_to_register, _serial_typename=None):
        '''register cls_to_register as a serializable subclass of XarrayIoSerializable.
        _serial_typename: str or None
            name to use for serialization. if None, use cls_to_register.__name__.
        '''
        cls = cls_to_register
        if _serial_typename is None: _serial_typename = cls.__name__
        # ensure uniqueness:
        if _serial_typename in XarrayIoSerializable.SERIAL_TYPES:
            errmsg = (f"cannot register {cls} as {_serial_typename!r} because that name is already in use.")
            raise KeyError(errmsg)
        # register cls!
        cls._serial_typename = _serial_typename
        XarrayIoSerializable.SERIAL_TYPES[cls._serial_typename] = cls

    def serialize(self):
        '''return dict of info about self, including 'typename' key.
        '''
        typename = self._serial_typename
        d = self.to_dict()
        return {"typename": typename, **d}

    @staticmethod
    def deserialize(serial):
        '''creates an XarrayIoSerializable object from a serialized representation.

        serial: dict or str
            str --> convert to dict via ast.literal_eval.
            dict must contain 'typename' key.

        The output here will be an instance of XarrayIoSerializable.SERIAL_TYPES[typename]
        '''
        _serial_input = serial  # for reference. Helps with debugging.
        if isinstance(serial, str):
            serial = ast.literal_eval(serial)
        serial = serial.copy()
        try:
            typename = serial.pop('typename')
        except KeyError:
            errmsg = (f"expected 'typename' key in input serial, but got keys={serial.keys()}.\n"
                      "to proceed without typename, consider using cls.from_dict.")
            raise DeserializingError(errmsg) from None
        cls_to_make = XarrayIoSerializable.SERIAL_TYPES[typename]
        return cls_to_make.from_dict(serial)


''' --------------------- io tools --------------------- '''

def _xarray_io_filepaths(filename, *, exist_ok=False):
    '''return dict of strs for folder, basename (without extension), netcdf_file, text_file, notes_file.
    folder will be abspath of filename, with '.pcxarr' added if needed.
    exist_ok: bool. If False and folder already exists, crash with FileExistsError.
    '''
    if not filename.endswith('.pcxarr'):
        filename += '.pcxarr'
    filepath_exists = os.path.exists(filename)
    if filepath_exists and not exist_ok:
        raise FileExistsError(f"Directory {filename!r} already exists; use a different name or set exist_ok=True.")
    folder = os.path.abspath(filename)
    basename = os.path.splitext(os.path.basename(folder))[0]
    netcdf_file = os.path.join(folder, f'{basename}.nc')
    text_file = os.path.join(folder, f'{basename}.txt')
    notes_file = os.path.join(folder, f'notes.txt')
    return dict(folder=folder, basename=basename, netcdf_file=netcdf_file, text_file=text_file, notes_file=notes_file)

def _xarray_coord_serializations(array):
    '''return serializations dict, for all array coords made of XarrayIoSerializable objects.'''
    result = dict()
    for cname, carr in array.coords.items():  # carr is the array of value(s) of this coord.
        xx = carr.values  # xx is carr as a numpy array.
        if isinstance(xx.flat[0], XarrayIoSerializable):  # [TODO] xx with mixed types with xx[0] not serializable.
            # will need to save these coords to text file via serializing, not to netcdf.
            if xx.ndim == 0:
                result[cname] = xx.flat[0].serialize()
            elif xx.ndim == 1:
                result[cname] = [x.serialize() for x in xx]  # [TODO] DimensionValueList(xx).serialize() instead?
            else:
                errmsg = (f"serialization of 2D+ XarrayIoSerializable coords not yet implemented."
                         f"(cname={cname!r}; ndim={xx.ndim}.)")
                raise SerializingError(errmsg)
    return result

def _xarray_coord_deserializations(serial):
    '''return dict of coords, deserialized (result has XarrayIoSerializable objects).
    serial: dict of {cname: cserial}
        where cserial is a dict for 0d coord, or list for 1d coord.
        2d+ coords not yet implemented.
    '''
    coords = dict()
    for cname, cserial in serial.items():
        if isinstance(cserial, dict):  # 0d coord
            coords[cname] = XarrayIoSerializable.deserialize(cserial)
        elif isinstance(cserial, list):  # 1d coord
            coords[cname] = [XarrayIoSerializable.deserialize(x) for x in cserial]
        else:
            raise DeserializingError(f"unexpected serialization type for coord {cname!r}: {type(cserial)}")
    return coords

def _xarray_best_save_engine():
    '''returns the best engine (as a str) to use for xarray save operation.
    In order of preference: 'netcdf4' > 'h5netcdf' > 'scipy'.
        netcdf4 is only first for consistency with xarray preferences.
            But, it requires you separately install netcdf binary.
        h5netcdf is probably sufficient for most users,
            and you can can pip install it without the netcdf binary.
        scipy is okay, but only compatible with netcdf3, and doesn't allow compression;
            it's always better to use h5netcdf, unless you really can't stand
            the idea of doing "pip install h5netcdf" for some reason.
    '''
    try:
        import netCDF4
        return 'netcdf4'
    except ImportError:
        pass
    try:
        import h5netcdf
        return 'h5netcdf'
    except ImportError:
        pass
    try:
        import scipy
        return 'scipy'
    except ImportError:
        pass
    errmsg = ("No suitable engine found for xarray saving; checked netCDF4, h5netcdf, and scipy.\n"
              "Recommendation: pip install h5netcdf, then try again.")
    raise ImportError(errmsg)

def _xarray_engine_compression_defaults(engine):
    '''returns dict of default compression settings, or None if this engine can't do compression.
    engine: str ('netcdf4', 'h5netcdf', 'scipy')
    result will be {'zlib': True} for 'netcdf4' and 'h5netcdf', or None for 'scipy'.
    '''
    if engine == 'netcdf4' or engine == 'h5netcdf':
        return {'zlib': True}
    elif engine == 'scipy':
        return None

def _xarray_encoding_from_compress_dict(array, compress_dict):
    '''returns encoding dict for this array (or dataset), to apply compress_dict to every var.
    if DataArray, apply compress_dict to array.
    if Dataset, apply compress_dict to each data_var.
    '''
    # xarray.to_netcdf expects encoding = dict of {var: d} for each var,
    #    where d is a dict of encoding options, e.g. compress_dict.
    # in the DataArray case, var=array.name if not None and not a dimension's name,
    #    else __xarray_dataarray_variable__.
    # in the Dataset case, var will just be one of the data_vars' names
    if isinstance(array, xr.DataArray):
        if array.name is None or array.name in array.dims:
            var = '__xarray_dataarray_variable__'
        else:
            var = array.name
        result = {var: compress_dict}
    elif isinstance(array, xr.Dataset):
        result = {var: compress_dict for var in array.data_vars}
    else:
        raise InputError(f"unexpected array type {type(array)} in _xarray_encoding_from_compress_dict.")
    return result

def _to_netcdf_engine_and_encoding_kws(array, engine, compress, encoding):
    '''return dict of values for engine and encoding, based on inputs,
    converting defaults to usable values wherever values not provided explicitly.
    '''
    if compress is not None and encoding is not None:
        errmsg = ("cannot provide both compress and encoding; "
                  f"but got compress={compress!r} and encoding={encoding!r}.")
        raise InputConflictError(errmsg)
    if engine is None:
        engine = _xarray_best_save_engine()
    else:
        KNOWN_ENGINES = {'netcdf4', 'h5netcdf', 'scipy'}
        if engine not in KNOWN_ENGINES:
            raise InputError(f"Got engine={engine!r}; expected one of {KNOWN_ENGINES}.")
    if compress is None:
        compress = _xarray_engine_compression_defaults(engine)
        if compress is None:
            compress = False
    if compress == True:
        compress = _xarray_engine_compression_defaults(engine)
        if compress is None:
            raise InputConflictError(f"engine {engine!r} incompatible with compress=True")
    # <-- hopefully at this point, compress is a dict, or False.
    if compress != False:
        encoding = _xarray_encoding_from_compress_dict(array, compress)
    return dict(engine=engine, encoding=encoding)

def _xarray_reset_multi_indexes(array):
    '''return array with all MultiIndex indexes reset, and relevant info added to array.attrs.
    the relevant info will be added as array.assign_attrs(info), where
        info = {'reset_multi_index:{d}': (tuple of associated coords)) for each reset index d}
    '''
    reset_info = {}
    for (idx, _idict) in array.indexes.group_by_index():
        if isinstance(idx, pd.MultiIndex):
            reset_info[f'reset_multi_index:{idx.name}'] = tuple(idx.names)
            array = array.reset_index(idx.name)
    if reset_info:
        array = array.assign_attrs(reset_info)
    return array

def _xarray_restore_multi_indexes(array):
    '''return array with MultiIndex indexes restored, based on 'reset_multi_index:{d}' attrs.
    also pops 'reset_multi_index:{d}' from attrs.
    See _xarray_reset_multi_indexes for more details.
    '''
    array = array.copy()  # to avoid modifying original array.
    for key, names in list(array.attrs.items()):
        if key.startswith('reset_multi_index:'):
            #name = key[len('reset_multi_index:'):]
            array = array.set_xindex(names)  # xarray figures out the appropriate name automatically.
            array.attrs.pop(key)
    return array

def _xarray_expand_dict_attrs(array):
    '''return array with all dict-valued attrs expanded into separate attrs,
    like "__dictattr_{origattrname}__{dictkey}__".
    Crashes with SerializingError if any dict keys are not strings,
        or if any dict values are dicts (not yet implemented).

    E.g., if array.attrs['foo'] = {'a': 1, 'b': 2},
    then result will have attrs:
        '__dictattr_foo__a__' = 1
        '__dictattr_foo__b__' = 2
    '''
    array = array.copy()  # to avoid modifying original array.
    to_expand = [array]
    if isinstance(array, xr.Dataset):
        for var in array.data_vars:
            to_expand.append(array[var])
    for arr in to_expand:
        for aname, value in list(arr.attrs.items()):
            if isinstance(value, dict):
                for k, v in value.items():
                    if not isinstance(k, str):
                        errmsg = (f"cannot expand dict-valued attr {aname!r} with non-string key(s). "
                                  f"Got non-string key: {k!r}.")
                        raise SerializingError(errmsg)
                    if isinstance(v, dict):
                        errmsg = (f"Serializing dict-valued attr {aname!r} whose values are also dicts "
                                    f"is not yet implemented. Got dict at array[{aname!r}][{k!r}].")
                        raise SerializingError(errmsg)
                    arr.attrs[f'__dictattr_{aname}__{k}__'] = v
                arr.attrs.pop(aname)
    return array

def _xarray_unexpand_dict_attrs(array):
    '''return array with all expanded-dict-valued attrs unexpanded back into dicts.
    expanded-dict-valued attrs look like "__dictattr_{origattrname}__{dictkey}__".

    if array is a Dataset, apply this to top-level attrs AND attrs for each data_var.
    '''
    array = array.copy()  # to avoid modifying original array.
    to_unexpand = [array]
    if isinstance(array, xr.Dataset):
        for var in array.data_vars:
            to_unexpand.append(array[var])
    pattern = r'__dictattr_(.+)__(.+)__'
    for arr in to_unexpand:
        dictattrs = {}
        for aname, value in list(arr.attrs.items()):
            match = re.match(pattern, aname)
            if match:
                origattrname, dictkey = match.groups()
                dictattrs.setdefault(origattrname, {})[dictkey] = value
                arr.attrs.pop(aname)
        for origattrname, dictvalue in dictattrs.items():
            arr.attrs[origattrname] = dictvalue
    return array

def _xarray_stringify_bool_attrs(array):
    '''return array with all bool-valued attrs replaced with strings,
    like "__bool=True__" or "__bool=False__".
    (for whatever reason, netcdf doesn't like storing bool-valued attributes...)

    if array is a Dataset, apply this to top-level attrs AND attrs for each data_var.
    '''
    array = array.copy()  # to avoid modifying original array.
    to_stringify = [array]
    if isinstance(array, xr.Dataset):
        for var in array.data_vars:
            to_stringify.append(array[var])
    for arr in to_stringify:
        for key, value in list(arr.attrs.items()):
            if isinstance(value, bool):
                arr.attrs[key] = f"__bool={value}__"
    return array

def _xarray_unstringify_bool_attrs(array):
    '''return array with all stringified-bool-valued attrs replaced with bools.
    (for whatever reason, netcdf doesn't like storing bool-valued attributes...)
    stringified-bools are "__bool=True__" or "__bool=False__".

    if array is a Dataset, apply this to top-level attrs AND attrs for each data_var.
    '''
    array = array.copy()  # to avoid modifying original array.
    to_stringify = [array]
    if isinstance(array, xr.Dataset):
        for var in array.data_vars:
            to_stringify.append(array[var])
    for arr in to_stringify:
        for key, value in list(arr.attrs.items()):
            if isinstance(value, str):
                if value == '__bool=True__':
                    arr.attrs[key] = True
                elif value == '__bool=False__':
                    arr.attrs[key] = False
                else:  # keep other strings unchanged.
                    pass
    return array


''' --------------------- saving & loading --------------------- '''

def _xarray_save_prep(array, *, exist_ok=False, add_meta=True,
                      reset_multi_index=True, expand_dict_attrs=True, stringify_bool_attrs=True,
                      engine=None, compress=None, encoding=None, **kw_to_netcdf):
    '''return array (prepped for saving), serializations, and to_netcdf kwargs.
    provided as a separate function to help with debugging.
    '''
    # to_netcdf kwargs bookkeeping
    kw_to_netcdf.update(_to_netcdf_engine_and_encoding_kws(array, engine, compress, encoding))
    # add code meta attrs bookkeeping
    if add_meta: array = array.assign_attrs(code_snapshot_info())
    # serialize XarrayIoSerializable coords
    serializations = _xarray_coord_serializations(array)
    # reset multi indexes, if requested (and any multi indexes present)
    if reset_multi_index: array = _xarray_reset_multi_indexes(array)
    # drop serializations
    array = array.drop_vars(list(serializations.keys()))
    # expand dict-valued attrs into separate attrs
    if expand_dict_attrs: array = _xarray_expand_dict_attrs(array)
    # replace bool-valued attributes with strings like: "__bool=True__" or "__bool=False__"
    if stringify_bool_attrs: array = _xarray_stringify_bool_attrs(array)
    # convert to xarray.Dataset if needed
    return array, serializations, kw_to_netcdf

@pcAccessor.register('save')
def xarray_save(array, filename=None, *, exist_ok=False, add_meta=True, notes=None,
                reset_multi_index=True, expand_dict_attrs=True, stringify_bool_attrs=True,
                engine=None, compress=None, encoding=None, **kw_to_netcdf):
    '''saves the array or dataset as filename.nc with a companion text file filename.txt.
    Both will be saved into a new directory named filename.pcxarr
    ("pcxarr" stands for "PlasmaCalcs xarray.DataArray or xarray.Dataset object")

    array: xarray.DataArray or xarray.Dataset
        the array or dataset to save
    filename: None or str
        where to save the array. Extension ".pcxarr" will be added if not present.
        None --> infer filename=array.name, or "unnamed_array" if array.name is None.
                (actually: array.name.replace('/', '÷'). To avoid interpreting division as directories.)
        if filename implies directories, those directories will be created, as per os.makedirs.
    exist_ok: bool, default False
        whether it's okay if directory with the target name to already exist.
        False --> crash with FileExistsError if directory exists.
        True --> might overwrite files in that directory!
    add_meta: bool
        whether to array.assign_attrs(details about current version of PlasmaCalcs code)
        Those details include 'pc__version__', 'pc__commit_hash', and 'datetime'.
    notes: None or object to convert to str
        if provided, also save a notes.txt file containing str(notes).
        Feel free to use this to write anything, e.g. describe what array means, in words.
    reset_multi_index: bool
        whether to array.reset_index(idx) for all MultiIndex idx, before saving.
        if True, also add info to array.attrs['reset_multi_index:{d}' for reset d].
        if False, may crash with NotImplementedError if any MultiIndex present.
    expand_dict_attrs: bool
        whether to expand dict-valued attributes into separate attributes,
        with names like "__dictattr_{origattrname}__{dictkey}__".
        (netcdf doesn't allow saving dict-valued attributes.)
        (will be applied before stringify_bool_attrs.)
    stringify_bool_attrs: bool
        whether to replace bool-valued attributes' values with "__bool=True__" or "__bool=False__".
        (netcdf doesn't like saving bool-valued attributes.)
        (xarray_load will automatically convert these back into bools.)

    additional kwargs relate to internal strategy for to_netcdf:
        engine: None or str ('netcdf4', 'h5netcdf', 'scipy')
            which engine to use for saving.
            None --> use _xarray_best_save_engine()
                    (picks first available, from netcdf4 > h5netcdf > scipy)
        compress: None, bool, or dict
            whether to compress data when writing.
            None --> True if engine can do compression, else False.
                    ('scipy' engine is not compatible with compression.)
            bool --> get dict from _xarray_engine_compression_defaults(engine)
                    (crash with InputConflictError if engine can't do compression.)
                    (default {'zlib': True} for 'netcdf4' and 'h5netcdf')
            dict --> apply this strategy to each data variable.
                    equivalent: encoding={var1: compress, var2: compress, ...}
        encoding: None or dict
            dict of {var: {encoding options for var}} across data vars.
            determined automatically if provided compress.

    returns abspath to filename.pcxarr directory where the array was saved.
    '''
    # filename/bookkeeping
    if filename is None:
        if array.name is None:
            filename = "unnamed_array"
        else:
            filename = array.name.replace('/', '÷')  # slashes in array name represent division, not file paths!
    paths = _xarray_io_filepaths(filename, exist_ok=exist_ok)
    # prepare array for saving. Also computes serializations.
    kw_prep = dict(exist_ok=exist_ok, add_meta=add_meta,
                   reset_multi_index=reset_multi_index, expand_dict_attrs=expand_dict_attrs,
                   stringify_bool_attrs=stringify_bool_attrs,
                   engine=engine, compress=compress, encoding=encoding, **kw_to_netcdf)
    array, serializations, kw_to_netcdf = _xarray_save_prep(array, **kw_prep)
    # make folder (if needed) and save files. First, array without serialized coords:
    os.makedirs(paths['folder'], exist_ok=True)
    if os.path.exists(paths['netcdf_file']):
        os.remove(paths['netcdf_file'])  # (some engines require to delete old file first)
    array.to_netcdf(paths['netcdf_file'], **kw_to_netcdf)
    # save serialized coords to text file. (also, first, set xarray_object_type='DataArray' or 'Dataset')
    serializations['xarray_object_type'] = type(array).__name__
    with open(paths['text_file'], 'w') as f:
        print(serializations, file=f)  # write to file
    # save notes to text file, if relevant.
    if notes is not None:
        with open(paths['notes_file'], 'w') as f:
            print(notes, file=f)
    # return abspath to directory where array was saved
    return paths['folder']

def xarray_load(filename, *, restore_multi_index=True, unstringify_bool_attrs=True,
                unexpand_dict_attrs=True, **kw__xarray_open):
    '''load the array or dataset from filename.pcxarr.
    filename: str
        where to load the array from. Extension ".pcxarr" will be added if not present.
    restore_multi_index: bool
        whether to restore MultiIndexes if 'reset_multi_index:{d}' in attrs (for some d)
    unstringify_bool_attrs: bool
        whether to convert stringified-bool-valued attributes back into bools.
        (stringified-bool-valued attributes have value "__bool=True__" or "__bool=False__")
    unexpand_dict_attrs: bool
        whether to unexpand expanded-dict-valued attrs attributes back into dicts.
        (expanded-dict-valued attrs look like "__dictattr_{origattrname}__{dictkey}__")
        (will be applied after unstringify_bool_attrs.)

    additional kwargs go to xarray.open_dataarray, or xarray.open_dataset
    
    if 'filepath' not in result.attrs, add 'filepath': abspath(filename).
    '''
    # filename/bookkeeping
    paths = _xarray_io_filepaths(filename, exist_ok=True)
    # load xarray type & coords info from text file then deserialize:
    with open(paths['text_file'], 'r') as f:
        serial_str = f.read()
    serial = ast.literal_eval(serial_str)
    xarray_object_type = serial.pop('xarray_object_type', 'DataArray')  # default DataArray for backwards compatibility
    # deserialize coords (do this first in case of crash to avoid loading lots of data before crashing.)
    coords = _xarray_coord_deserializations(serial)
    # load array or dataset from netcdf file:
    if xarray_object_type == 'DataArray':
        array = xr.open_dataarray(paths['netcdf_file'], **kw__xarray_open)
    elif xarray_object_type == 'Dataset':
        array = xr.open_dataset(paths['netcdf_file'], **kw__xarray_open)
    else:
        raise DeserializingError(f"unexpected xarray_object_type {xarray_object_type!r}, from {text_filename!r}.")
    # assign coords to array or dataset:
    array = array.assign_coords(coords)
    # restore MultiIndexes if requested:
    if restore_multi_index: array = _xarray_restore_multi_indexes(array)
    # unstringify bool-valued attributes if requested:
    if unstringify_bool_attrs: array = _xarray_unstringify_bool_attrs(array)
    # unexpand dict-valued attrs if requested:
    if unexpand_dict_attrs: array = _xarray_unexpand_dict_attrs(array)
    # assign 'filepath' to attrs if not already present:
    if 'filepath' not in array.attrs:
        array = array.assign_attrs({'filepath': paths['folder']})
    return array

def xarray_mergeload(dirpath, **kw__xarray_load):
    '''xarray_load all .pcxarr files from dirpath, then xr.merge() the results.
    kwargs will be passed to xarray_load.
    '''
    # dirpath/bookkeeping
    dirpath = os.path.abspath(dirpath)
    if not os.path.isdir(dirpath):
        raise FileNotFoundError(f"the dirpath directory does not exist: dirpath={dirpath!r}")
    # find all .pcxarr files in dirpath:
    files = [f for f in os.listdir(dirpath) if f.endswith('.pcxarr')]
    if len(files)==0:
        raise FileNotFoundError(f"the dirpath directory has no .pcxarr files: dirpath={dirpath!r}.")
    files = sorted(files)  # sort to ensure consistent order. Probably doesn't affect merge results...
    # load all .pcxarr files:
    arrays = []
    for f in files:
        filepath = os.path.join(dirpath, f)
        arrays.append(xarray_load(filepath, **kw__xarray_load))
    # merge all arrays:
    if len(arrays) == 1:
        return arrays[0]
    else:
        return xr.merge(arrays)



"""
File Purpose: tools for iterables
"""
import operator

import numpy as np

from . import properties  # for unambiguous "alias" inside DictWithAliases class.
from .properties import alias, alias_child, simple_property
from .sentinels import UNSET
from .timing import ProgressUpdater
from ..errors import InputError, InputMissingError, InputConflictError, DimensionalityError
from ..defaults import DEFAULTS


''' ----------------------------- Misc ----------------------------- '''

def is_iterable(x):
    '''returns True if x is iterable, False otherwise.'''
    try:
        iter(x)
        return True
    except TypeError:
        return False

def is_dictlike(x):
    '''returns True if x is dict-like, False otherwise.
    returns x.is_dictlike if it exists,
    else returns whether x has keys() and __getitem__.
    '''
    try:
        x_is_dictlike = x.is_dictlike
    except AttributeError:
        return hasattr(x, 'keys') and hasattr(x, '__getitem__')
    else:
        return x_is_dictlike()

def argmax(iterable):
    '''returns first occurence of maximum value in iterable.
    converts iterable to tuple before beginning search.
    [EFF] might not be efficient for large iterables, but it's probably fine.
    '''
    l = tuple(iterable)
    return l.index(max(l))

def rargmax(iterable):
    '''returns first occurence of maximum value in iterable, starting search from the end.
    converts iterable to tuple before beginning search.
    [EFF] might not be efficient for large iterables, but it's probably fine.
    '''
    l = tuple(iterable)
    return len(l) - 1 - l[::-1].index(max(l))

def scalar_item(arraylike):
    '''returns the single item from arraylike. raise InputError if not possible.
    if arraylike.ndim or .size exists:
        return arraylike.item() if ndim==0 or size==1, else crash.
    else:
        return arraylike[0] if len(arraylike)==1, else crash.
    '''
    if hasattr(arraylike, 'ndim') or hasattr(arraylike, 'size'):
        if getattr(arraylike, 'ndim', None) == 0 or getattr(arraylike, 'size', None) == 1:
            return arraylike.item()
        else:
            errmsg = f'expected array with ndim==0 or size==1, but got ndim={ndim} and size={size}'
            raise InputError(errmsg)
    else:
        if len(arraylike) == 1:
            return arraylike[0]
        else:
            errmsg = f'expected iterable with len==1, but got len={len(arraylike)}'
            raise InputError(errmsg)


''' --------------------- dict with aliases --------------------- '''

class DictWithAliases(dict):
    '''dict but some keys are aliases for other keys.
    E.g. d = AliasDict({'x': 7}); d.alias('x', 'y');
        d['y'] == d['x'] == 7
        d['y'] = 8
        d['y'] == d['x'] == 8

    Checking "key in self" also checks aliases.
        (use "key in self.keys()" to exclude aliases.)
    Looping through self.keys() or self.items() does NOT include aliases.

    add alias via self.alias(key0, key1),
        and/or provide dict of aliases during __init__.
    '''
    DEFAULT_ALIASES = {}  # dict of {alias: official key}

    aliases = simple_property('_aliases', setdefaultvia='_default_aliases', setable=False,
        doc='''aliases known by this dict. aliases is a dict of {alias: key}''')
        # aliases.keys() must be the aliases because each key can have multiple aliases.

    def _default_aliases(self):
        '''returns default aliases for this dict.
        the implementation here just returns self.DEFAULT_ALIASES.copy().
        '''
        return self.DEFAULT_ALIASES.copy()

    def __init__(self, *args_dict, aliases=UNSET, **kwargs_dict):
        '''initialize self with key-value pairs, and optionally aliases.
        aliases: UNSET, or dict of {key1: key0} pairs to be aliased.
            UNSET --> self will use self.DEFAULT_ALIASES.
            dict --> self.alias(key0, key1) for each (key0, key1) pair.
                    (key1 is more likely to become the "alias", internally.)
        '''
        super().__init__(*args_dict, **kwargs_dict)
        if aliases is not UNSET:
            self.add_aliases(aliases)

    def __getitem__(self, key):
        return super().__getitem__(self.aliases.get(key, key))

    def __setitem__(self, key, value):
        return super().__setitem__(self.aliases.get(key, key), value)

    def __delitem__(self, key):
        '''deletes value associated with key (or alias) from self.
        does NOT alter anything about known aliases.
        '''
        return super().__delitem__(self.aliases.get(key, key))

    def pop(self, key, default=UNSET):
        '''pop key from self, returning self[key].
        does NOT alter anything about known aliases.

        default: UNSET or any value
            if key not in self, return default if provided (i.e., not UNSET)
        '''
        if default is UNSET:
            return super().pop(self.aliases.get(key, key))
        else:
            return super().pop(self.aliases.get(key, key), default)

    def update(self, *args_dict, **kwargs_dict):
        '''update self with new key-value pairs.
        does NOT alter anything about known aliases.
        handles appropriately any known aliases in key implied by *args and **kwargs.
        '''
        for k, v in dict(*args_dict, **kwargs_dict).items():
            self[self.aliases.get(k, k)] = v

    def copy(self):
        '''return a shallow copy of self. Result aliases are the same as self.aliases.'''
        return type(self)(self)

    def __contains__(self, key):
        return super().__contains__(self.aliases.get(key, key))

    @property
    def _known_nonalias_keys(self):
        '''all keys known by self as "official" keys (not aliases).
        (official keys and aliases behave the same way - user shouldn't need to worry about this.)
        set of all self.keys() plus all values in self.aliases.
        '''
        return set(self.keys()) | set(self.aliases.values())

    def alias(self, key0, key1):
        '''tell self to treat these keys as the same key.
        key0 is more likely to become the official key, while key1 will be the alias.
        (however, if either is already known to self, decides "official vs alias" appropriately.)
        '''
        # decide which is the "official" key (in self.keys()) and which is the alias.
        if key0 == key1:
            raise InputConflictError(f'cannot alias key to itself: key={key0!r}')
        known_nonalias_keys = self._known_nonalias_keys
        if key0 in known_nonalias_keys:
            official, alias = key0, key1
        elif key1 in known_nonalias_keys:
            official, alias = key1, key0
        else:  # we have no prior info about these keys. Use default: official = key0.
            official, alias = key0, key1
        # check for conflict with existing keys
        if official in self.keys() and alias in self.keys():
            raise InputError( f'cannot alias({key0!r}, {key1!r}); both keys already in self.keys()!')
        # check for conflict with existing aliases.
        if alias in self.aliases:
            if self.aliases[alias] != official:
                errmsg = (f'alias({key0!r}, {key1!r}) conflicts with existing alias: '
                         f'{alias!r} <--> {self.aliases[alias]!r}.'
                         f'\nConsider using self.un_alias to remove a known alias.')
                raise InputError(errmsg)
        # done!
        self.aliases[alias] = official

    add_alias = properties.alias('alias')  # a perfectly sensible line of code :)

    def add_aliases(self, aliases):
        '''add multiple aliases at once.
        aliases: dict of {alias: key} pairs.

        Equivalent to calling self.alias(key, alias) for each (alias, key) pair.
        '''
        for alias, key in aliases.items():
            self.alias(key, alias)

    def un_alias(self, alias):
        '''remove alias from self.
        If alias is an "official" key with a known alias,
            swap the alias for the official key, then remove the previously-official key.
        '''
        raise NotImplementedError('[TODO] DictWithAliases.un_alias')

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [super().__repr__()]
        if len(self.aliases)>0:
            contents.append(f'aliases={self.aliases}')
        contents_str = f",\n{DEFAULTS.TAB}".join(contents)
        return f'{type(self).__name__}({contents_str})'


''' ----------------------------- custom Dictlike ----------------------------- '''

class DictlikeFromKeysAndGetitem():
    '''Dict-like object, assuming keys() and __getitem__ are defined.'''
    def __getitem__(self, key):
        '''get value for this key'''
        raise NotImplementedError(f'{type(self).__name__}.__getitem__')

    def keys(self):
        '''return tuple of keys in self.'''
        raise NotImplementedError(f'{type(self).__name__}.keys()')

    def __iter__(self):
        '''raises TypeError; use .keys(), .values(), or .items() instead.'''
        raise TypeError(f'{type(self).__name__} is not iterable; use .keys(), .values(), or .items() instead.')

    def values(self):
        '''return tuple of values corresponding to self.keys().'''
        return tuple(self[key] for key in self.keys())

    def items(self):
        '''return tuple of (key, value) pairs corresponding to self.keys() and self.values().'''
        return tuple(zip(self.keys(), self.values()))

    def get(self, key, default=UNSET):
        '''return self[key]. if default is provided and self[key] doesn't exist, return default.'''
        try:
            return self[key]
        except KeyError:
            if default is UNSET:
                raise
            return default


''' ----------------------------- partitioning stuff ----------------------------- '''

class Partition(DictlikeFromKeysAndGetitem):
    '''(ordered) dict of lists partitioned by category.
    (assumes python >= 3.7, i.e. dict order is guaranteed to be insertion order)
    self assumes it will remain unchanged after originally being created.

    self.partition stores {key: list of elements in category}
    self.idx stores {key: list of indices of elements in original}
    self.flat stores all elements in category order, in a single list.
    self.idx_flat concatenates self.idx.values().
    self.ridx stores {key: [i such that self.flat[i] == x for x in this category]}
    self.ridx_flat stores [i such that self.flat[i] == x for x in original]

    iterating through self yields (key, value) pairs.

    original: list-like
        the original iterable, to be partitioned.
    f: callable
        maps original elements to category keys. categories must be hashable.
        Will call f(x) for x in original, to determine categories.
    keys: ordered list of categories
        use these as known categories if provided.
    '''
    def __init__(self, original, f, *, keys=[]):
        self.original = original
        self.f = f
        self._init_keys = keys
        self.init_all()

    def init_all(self):
        '''init all the things, based on self.original, self.f, and self._init_keys'''
        self.init_partition()
        self.init_flat()
        self.init_ridx()

    def init_partition(self):
        '''creates self.partition and self.idx'''
        f = self.f
        partition = dict()
        idx = dict()
        for key in getattr(self, '_init_keys', []):
            partition[key] = []
            idx[key] = []
        for i, x in enumerate(self.original):
            key = f(x)
            partition.setdefault(key, []).append(x)
            idx.setdefault(key, []).append(i)
        self.partition = partition
        self.idx = idx

    def init_flat(self):
        '''create self.flat and self.idx_flat.
        self.flat stores all elements in order of categories, in a single list.
        self.idx_flat stores the indices of all elements in self.flat.
        '''
        flat = []
        for key in self.keys():
            flat.extend(self.partition[key])
        self.flat = flat
        idx_flat = []
        for key in self.keys():
            idx_flat.extend(self.idx[key])
        self.idx_flat = idx_flat

    def init_ridx(self):
        '''create self.ridx and self.ridx_flat.
        self.ridx stores {key: [i such that self.flat[i] == x for x in this category]}
        self.ridx_flat stores [i such that self.flat[i] == x for x in original]
        '''
        ridx_flat = []
        for i, x in enumerate(self.flat):
            ridx_flat.append(self.idx_flat.index(i))
        self.ridx_flat = ridx_flat
        ridx = dict()
        istart = 0
        for key in self.keys():
            cat_len = len(self.partition[key])
            ridx[key] = ridx_flat[istart:istart+cat_len]
            istart = istart + cat_len
        self.ridx = ridx

    def __getitem__(self, key):
        return self.partition[key]

    def keys(self):
        return list(self.partition.keys())

    # intentionally does not define __setitem__; expects to remain unchanged after __init__.

    def __iter__(self):
        '''iterate through (key, val) pairs.'''
        return iter(self.partition.items())

    def __contains__(self, key):
        '''tells whether key is in self.partition'''
        return key in self.partition

    # # # DISPLAY # # #
    def __repr__(self):
        return f'{type(self).__name__}({self.partition})'


class PartitionFromXarray(Partition):
    '''(ordered) dict of lists partitioned by category.
    (assumes python >= 3.7, i.e. dict order is guaranteed to be insertion order)
    self assumes it will remain unchanged after originally being created.

    self.partition stores {key: list of elements in category}
    self.idx stores {key: list of indices of elements in original}
    self.flat stores all elements in category order, in a single list.
    self.idx_flat concatenates self.idx.values().
    self.ridx stores {key: [i such that self.flat[i] == x for x in this category]}
    self.ridx_flat stores [i such that self.flat[i] == x for x in original]

    iterating through self yields (key, value) pairs.

    xarray: xarray.DataArray with ndim <= 1
        data array determining partition category for o in original.
        self.partition keys will be values of xarray;
        partition values will be lists of o from original for which xarray[coord] == o.
        E.g. xr.DataArray(['a', 'a', 'b', 'a', 'c'], coords={coord: ['o1', 'o2', 'o3', 'o4', 'o5']})
            --> self.partition = {'a': ['o1', 'o2', 'o4'], 'b': ['o3'], 'c': ['o5']}
    coord: None or str
        coordinate of xarray containing all values from original.
        None --> infer as coord = xarray.dims[0] if ndim=1,
                 else xarray.coords[0] if ndim=0 and len(xarray.coords)==1,
                 else crash.
    original: None or list-like
        the original iterable, to be partitioned.
        None --> use list(xarray[coord].values).
    '''
    def __init__(self, xarray, coord=None, original=None):
        if not hasattr(xarray, 'ndim'):
            raise InputError(f'expected xarray-like object, but obj does not even have ndim attr: obj={obj}')
        if xarray.ndim >= 2:
            raise DimensionalityError(f'PartitionFromXarray expects xarray.ndim<=1; got {xarray.ndim}.')
        self.xarray = xarray
        if coord is None:
            if xarray.ndim == 1:
                coord = xarray.dims[0]
            elif xarray.ndim == 0:
                if len(xarray.coords) == 1:
                    coord = list(xarray.coords)[0]
                else:
                    raise InputMissingError(f'coord, when xarray.ndim=0 and len(xarray.coords)>1')
            else:
                raise InputMissingError(f'coord, when xarray.ndim>1')
        if original is None:
            original = list(xarray[coord].values)
        self.original = original
        self.coord = coord
        self.init_all()

    def f(self, x):
        '''return category for x.'''
        # don't use xarray.sel; use == instead:
        cname = self.coord
        array = self.xarray
        cvals = array[cname].values
        if array.ndim == 0:
            if x == cvals.item():
                return array.item()
        else:
            for i, c in enumerate(cvals):
                if x == c:
                    return array.isel({cname: i}).item()
        raise InputConflictError(f'x={x!r} not found in self.xarray[{cname!r}]; cannot determine f(x)')


''' ----------------------------- Generic Containers ----------------------------- '''

class Container():
    '''a container for multiple objects, & rules for enumerating & indexing.
    Here, implements self.__getitem__ so that self[i]=self.data[i],
        and self.enumerate which yields (i, self[i]) pairs.
    subclass should implement __init__, _enumerate_all, and new_empty.
    '''
    # # # THINGS THAT SUBCLASS SHOULD IMPLEMENT # # #
    def __init__(self, stuff):
        '''should set self.data = something related to stuff.'''
        raise NotImplementedError(f'{self.__class__.__name__}.__init__')

    def _enumerate_all(self):
        '''iterate through all objs in self, yielding (i, self[i]) pairs.
        Equivalent to self.enumerate(idx=None).
        The implementation will depend on the container type; subclass should implement.
        '''
        raise NotImplementedError(f'{self.__class__.__name__}._enumerate_all')

    def new_empty(self, fill=UNSET):
        '''return a new container of the same shape as self, filled with the value fill.
        The implementation will depend on the container type; subclass should implement.
        '''
        raise NotImplementedError(f'{self.__class__.__name__}.new_empty')

    def _size_all(self):
        '''return the number of objects in the container.
        The implementation will depend on the container type; subclass should implement.
        '''
        raise NotImplementedError(f'{self.__class__.__name__}.size_all')

    # # # GETITEM & ENUMERATE # # #
    def __getitem__(self, idx):
        return self.data[idx]

    def enumerate(self, idx=None):
        '''iterate through i in idx, yielding (i, self[i]) pairs.
        If idx is None, iterate through all objs in self (see self._enumerate_all).
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        if idx is None:
            for i, selfi in self._enumerate_all():
                yield (i, selfi)
        else:
            for i in idx:
                yield (i, self[i])

    def size(self, idx=None):
        '''return the number of objects in the container, or in idx if provided.'''
        if idx is None:
            return self._size_all()
        return len(idx)

    # # # DISPLAY # # #
    def __repr__(self):
        return f'{self.__class__.__name__}({self.data})'


class ContainerOfList(Container):
    '''a list-like container.'''
    def __init__(self, objs):
        self.data = [o for o in objs]

    def _enumerate_all(self):
        '''iterate through all objs in self, yielding (i, self[i]) pairs.'''
        return enumerate(self.data)

    def new_empty(self, fill=UNSET):
        '''return a new list of the same shape as self, filled with the value fill.'''
        return [fill for _ in self.data]

    def _size_all(self):
        '''the number of objects in this container. == len(self.data)'''
        return len(self.data)


class ContainerOfArray(Container):
    '''a numpy-array-like container.'''
    def __init__(self, arr):
        self.data = np.asanyarray(arr)

    def _enumerate_all(self):
        '''iterate through all objs in self, yielding (i, self[i]) pairs.'''
        return np.ndenumerate(self.data)

    def new_empty(self, fill=UNSET):
        '''return a new array of the same shape as self, filled with the value fill.'''
        return np.full_like(self.data, fill, dtype=object)

    def _size_all(self):
        '''the number of objects in this container. == self.data.size'''
        return self.data.size

    shape = alias_child('data', 'shape')
    ndim = alias_child('data', 'ndim')
    dtype = alias_child('data', 'dtype')


class ContainerOfDict(Container):
    '''a dict-like container.'''
    def __init__(self, d):
        self.data = dict(d)  # copy the dict (and ensure dict-like)

    def _enumerate_all(self):
        '''iterate through all objs in self, yielding (i, self[i]) pairs.'''
        return self.data.items()

    def new_empty(self, fill=UNSET):
        '''return a new dict of the same shape as self, filled with the value fill.'''
        return {k: fill for k in self.data.keys()}

    def _size_all(self):
        '''the number of objects in this container. == len(self.data)'''
        return len(self.data)


''' ----------------------------- Bijection ----------------------------- '''

class Bijection(dict):
    '''stores forward and backward mapping.
    behaves like forward mapping, but self.inverse is the reverse mapping.
    '''
    inverse = simple_property('_inverse', setdefault=dict,
        doc='''reverse mapping. {value: key} for all items in this bijection.''')

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.inverse[value] = key

    def __getitem__(self, key):
        result = super().__getitem__(key)
        return result

    def __delitem__(self, key):
        value = self.pop(key)

    def clear(self):
        '''clear all items from this bijection.'''
        super().clear()
        self.inverse.clear()

    def pop(self, key, default=UNSET):
        '''pop the value for this key, and return it.'''
        value = super().pop(key, default)
        if value is not UNSET:
            del self.inverse[value]
        return value

    def popitem(self):
        '''pop an item from this bijection, and return it.'''
        key, value = super().popitem()
        del self.inverse[value]
        return key, value

    def update(self, *args, **kw):
        '''update this bijection with new items.'''
        super().update(*args, **kw)
        self.inverse.update({v: k for k, v in self.items()})

    def __repr__(self):
        return f'{type(self).__name__}({super().__repr__()})'


class BijectiveMemory(Bijection):
    '''bijection which also stores the next key to use. Keys should be numbers.
    self.nextkey is never decremented (don't "fill in gaps" after deleting keys)
    '''
    nextkey = simple_property('_nextkey', setdefaultvia='_default_nextkey',
        doc='''next key for self.key(value) if new value.
        nextkey >= max(self.keys(), default=-1) + 1.''')
    def _default_nextkey(self):
        return max(self.keys(), default=-1) + 1

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.nextkey = max(self.nextkey, key) + 1

    def update(self, *args, **kw):
        super().update(*args, **kw)
        self.nextkey = max([self.nextkey-1, *self.keys()]) + 1

    def key(self, value):
        '''return the key associated with this value; store the value if not already stored.'''
        if value in self.inverse:
            return self.inverse[value]
        else:
            key = self.nextkey
            self[key] = value
            return key


''' ----------------------------- SymmetricPairMapping ----------------------------- '''

class SymmetricPairMapping():
    '''mapping with keys pairs, where order of the pair doesn't matter.

    mapping: dict of {(obj1, obj2): value}

    The keys (obj2, obj1) and (obj1, obj2) will be treated the same way,
        when getting/setting/deleting items, and when checking if key in self.
        [EFF] For efficiency, only stores one of these pairs (the one where obj1 <= obj2).
    '''
    def __init__(self, mapping=dict()):
        self.mapping = dict()
        for key, value in mapping.items():
            self[key] = value

    def _sanitize_key(self, key):
        '''return sanitized key (obj1, obj2) such that obj1 <= obj2.'''
        if not isinstance(key, tuple) or len(key) != 2:
            raise TypeError(f'key={key!r} must be a tuple of 2 objs')
        if key[1] < key[0]:
            key = (key[1], key[0])
        return key
    
    def __setitem__(self, key, value):
        key = self._sanitize_key(key)
        self.mapping[key] = value

    def __getitem__(self, key):
        key = self._sanitize_key(key)
        return self.mapping[key]

    def __delitem__(self, key):
        key = self._sanitize_key(key)
        del self.mapping[key]

    def __contains__(self, key):
        key = self._sanitize_key(key)
        return key in self.mapping

    def update(self, other):
        '''update self with other.
        other: dict of {(obj1, obj2): value}
        '''
        for key, value in other.items():
            self[key] = value

    def clear(self):
        '''remove all known items from self.'''
        self.mapping.clear()

    def pop(self, key, default):
        '''remove self[key], and return its value. if key not in self, return default.'''
        key = self._sanitize_key(key)
        return self.mapping.pop(key, default)

    def __len__(self):
        return len(self.mapping)
    def values(self):
        return self.mapping.values()
    def items(self):
        return self.mapping.items()
    def keys(self):
        return self.mapping.keys()

    def __iter__(self):
        '''iterate over self.items (not self.keys).'''
        return iter(self.items())

    def copy(self):
        '''return a shallow copy of self.'''
        return type(self)(self.mapping)

    def __eq__(self, other):
        if not isinstance(other, SymmetricPairMapping):
            return False
        return self.mapping == other.mapping

    # # # DISPLAY # # #
    _repr_newlines = False  # whether to put newlines between keys in repr.

    def __repr__(self):
        if self._repr_newlines:
            SEP = f'\n{DEFAULTS.TAB}'
            JOINSEP = f',{SEP}'
            contents = [f'{k}: {v!r}' for k, v in self.mapping.items()]
            STARTSEP = SEP if len(contents) > 1 else ''
            ENDSEP = f'\n' if len(contents) > 3 else ''
            return f'{type(self).__name__}({STARTSEP}{JOINSEP.join(contents)}{ENDSEP})'
        else:
            return f'{type(self).__name__}({self.mapping!r})'


''' ----------------------------- DictOfSimilar ----------------------------- '''

class DictOfSimilar(dict):
    '''dict of similar objects with similar attributes.
    similar attributes list stored in SIMILAR_ATTRS.

    In many ways, simply broadcasts operations to all values in the dict.

    See self.to_ds(), to_da(), and to_xr() for helpful methods to convert result to xarray objects
        (These are defined in xarray_tools and bound to the DictOfSimilar class there).
    '''
    REPR_ITEM_MAXLEN = 50  # max length for repr for a single item; abbreviate if longer.
    SIMILAR_ATTRS = []
    cls_new = None  # if provided, use this to create new objects instead of type(self).

    def _new(self, *args, **kw):
        '''return a new DictOfSimilar of the same type as self.'''
        cls = type(self) if self.cls_new is None else self.cls_new
        return cls(*args, **kw)

    def apply(self, func, *args, **kw):
        '''return func(obj, *args, **kw) for each object in self.'''
        result = dict()
        for k, obj in self.items():
            result[k] = func(obj, *args, **kw)
        return self._new(result)

    do = alias('apply')

    gi = alias('getitems')
    si = alias('setitems')
    ga = alias('getattrs')
    sa = alias('setattrs')
    ca = alias('callattrs')

    def getitems(self, i):
        '''get obj[i] for each object in self.'''
        result = dict()
        for k, obj in self.items():
            result[k] = obj[i]
        return self._new(result)

    def setitems(self, i, value):
        '''set obj[i] = value for each object in self.'''
        for obj in self.values():
            obj[i] = value

    def getattrs(self, attr, default=UNSET):
        '''get obj.attr for each object in self.'''
        result = dict()
        for k, obj in self.items():
            if default is UNSET:
                result[k] = getattr(obj, attr)
            else:
                result[k] = getattr(obj, attr, default)
        return self._new(result)

    def setattrs(self, attr, value):
        '''set obj.attr = value for each object in self.'''
        for obj in self.values():
            setattr(obj, attr, value)

    def callattrs(self, attr, *args, dos_verbose=False, **kw):
        '''call obj.attr(*args, **kw) for each object in self.
        dos_verbose: bool or number
            whether to print progress updates from DictOfSimilar while looping.
            False or <=0 --> no updates
            True --> updates every 5 seconds
            >=2 and <5 --> updates every 2 seconds
            >=5 --> updates every loop iteration
        '''
        result = dict()
        updater = self._updater(dos_verbose)
        for i, (k, obj) in enumerate(self.items()):
            updater.print(f'Calling {i+1:2d} of {len(self)}: {k!r}')
            result[k] = getattr(obj, attr)(*args, **kw)
        updater.finalize(f'<{type(self).__name__} at {hex(id(self))}>.callattrs({attr!r})', always=False)
        return self._new(result)

    def __call__(self, *args, dos_verbose=True, **kw):
        '''call obj(*args, **kw) for each object in self.
        dos_verbose: bool or number
            whether to print progress updates from DictOfSimilar while looping.
            False or <=0 --> no updates
            True --> updates every 5 seconds
            >=2 and <5 --> updates every 2 seconds
            >=5 --> updates every loop iteration
        '''
        result = dict()
        updater = self._updater(dos_verbose)
        for i, (k, obj) in enumerate(self.items()):
            updater.print(f'Calling {i+1:2d} of {len(self)}: {k!r}')
            result[k] = obj(*args, **kw)
        updater.finalize(f'<{type(self).__name__} at {hex(id(self))}>.__call__', always=False)
        return self._new(result)

    def _updater(self, dos_verbose, *, wait=True):
        '''return ProgressUpdater with print_freq appropriate for dos_verbose.
        dos_verbose: bool or number
            whether to print progress updates from DictOfSimilar while looping.
            False or <=0 --> no updates
            True --> updates every 5 seconds
            >=2 and <5 --> updates every 2 seconds
            >=5 --> updates every loop iteration
        '''
        if (not dos_verbose) or dos_verbose <= 0:
            print_freq = -1
        elif dos_verbose >= 5:
            print_freq = 0
        elif dos_verbose >= 2:
            print_freq = 2
        else:
            print_freq = 5
        return ProgressUpdater(print_freq, wait=wait)

    # # # MISC. CONVENIENCE # # #
    def subset(self, keep=None, *, drop=None):
        '''return DictOfSimilar like self but with only a subset of the keys from self.
        
        keep: None, iterable, or callable
            None --> no restrictions from `keep`
            iterable --> only keep self[k] for k in keep.
            callable --> only keep (k,v) from self.items() if keep(k, v).
        drop: None, iterable, or callable
            None --> no restrictions from `drop`
            iterable --> drop self[k] for k in drop.
            callable --> drop (k,v) from self.items() if drop(k, v).
        If any conflict between drop and keep, drop takes priority.
        (E.g. if 'key1' in keep and in drop, it will be dropped, not kept.)
        '''
        if keep is None:
            keep = list(self.keys())
        elif callable(keep):
            keeping = []  # could do list comprehension, but loop makes debugging easier if crash.
            for k, v in self.items():
                if keep(k, v):
                    keeping.append(k)
            keep = keeping
        result = {k: self[k] for k in keep}
        if drop is None:
            drop = []
        elif callable(drop):
            dropping = []  # could do list comprehension, but loop makes debugging easier if crash.
            for k, v in result.items():
                if drop(k, v):
                    dropping.append(k)
            drop = dropping
        for k in drop:
            result.pop(k, None)
        try:
            _orig_cls_new = self.cls_new
            self.cls_new = None  # e.g., MultiCalculator self -> MultiCalculator result, not DictOfSimilar.
            result = self._new(result)
        finally:
            self.cls_new = _orig_cls_new
        return result

    # # # SIMILAR ATTRS BEHAVIOR # # #
    def __getattr__(self, attr):
        '''self.getattrs(attr) if attr in SIMILAR_ATTRS. Else raise AttributeError.
        (Only invoked when object.__getattr__(self, attr) fails.)
        '''
        if attr in self.SIMILAR_ATTRS:
            return self.getattrs(attr)
        else:
            raise AttributeError(f'{type(self).__name__} object has no attribute {attr!r}')

    def __setattr__(self, attr, value):
        '''self.setattrs(attr, value) if attr in SIMILAR_ATTRS, otherwise super().__setattr__.'''
        if attr in self.SIMILAR_ATTRS:
            return self.setattrs(attr, value)
        else:
            return super().__setattr__(attr, value)

    def __getitem__(self, i):
        '''self[i[0]].getitems(i[1]) if i tuple and i[0] in SIMILAR_ATTRS.
            (e.g. self['snap', 0] returns self['snap'].getitems(0)
        otherwise, try super().__getitem__.
        if that fails, try to return list(self.values())[i]
            (e.g., self[3] will return 3rd object, unless 3 in self.keys())
        '''
        if isinstance(i, tuple) and i[0] in self.SIMILAR_ATTRS:
            return self.getattrs(i[0]).getitems(i[1])
        else:
            try:
                return super().__getitem__(i)
            except KeyError:
                return list(self.values())[i]

    def __setitem__(self, i, value):
        '''self.setitems(i[1], value) if i tuple and i[0] in SIMILAR_ATTRS. otherwise super().__setitem__.'''
        if isinstance(i, tuple) and i[0] in self.SIMILAR_ATTRS:
            return self.getattrs(i[0]).setitems(i[1], value)
        else:
            return super().__setitem__(i, value)

    # # # ARITHMETIC APPLIES ACROSS ALL VALUES # # #
    def _math_op(self, other, op):
        '''apply op to self and other. If other is a dict, apply op to self[k] and other[k] for each k.
        Otherwise, apply op to each value in self and other.
        '''
        if is_dictlike(other):
            return self._new({k: op(self[k], other[k]) for k in self})
        else:
            return self.apply(op, other)

    def __add__(self, other):
        '''self + other. At each key if other is a dict, else apply to each value in self.'''
        return self._math_op(other, operator.add)

    def __radd__(self, other):
        '''other + self. At each key if other is a dict, else apply to each value in self.'''
        return self._math_op(other, lambda x, y: y + x)

    def __sub__(self, other):
        '''self - other. At each key if other is a dict, else apply to each value in self.'''
        return self._math_op(other, operator.sub)

    def __rsub__(self, other):
        '''other - self. At each key if other is a dict, else apply to each value in self.'''
        return self._math_op(other, lambda x, y: y - x)

    def __mul__(self, other):
        '''self * other. At each key if other is a dict, else apply to each value in self.'''
        return self._math_op(other, operator.mul)

    def __rmul__(self, other):
        '''other * self. At each key if other is a dict, else apply to each value in self.'''
        return self._math_op(other, lambda x, y: y * x)

    def __truediv__(self, other):
        '''self / other. At each key if other is a dict, else apply to each value in self.'''
        return self._math_op(other, operator.truediv)

    def __rtruediv__(self, other):
        '''other / self. At each key if other is a dict, else apply to each value in self.'''
        return self._math_op(other, lambda x, y: y / x)

    def __floordiv__(self, other):
        '''self // other. At each key if other is a dict, else apply to each value in self.'''
        return self._math_op(other, operator.floordiv)

    def __rfloordiv__(self, other):
        '''other // self. At each key if other is a dict, else apply to each value in self.'''
        return self._math_op(other, lambda x, y: y // x)

    def __mod__(self, other):
        '''self % other. At each key if other is a dict, else apply to each value in self.'''
        return self._math_op(other, operator.mod)

    def __rmod__(self, other):
        '''other % self. At each key if other is a dict, else apply to each value in self.'''
        return self._math_op(other, lambda x, y: y % x)

    def __pow__(self, other):
        '''self ** other. At each key if other is a dict, else apply to each value in self.'''
        return self._math_op(other, operator.pow)

    def __rpow__(self, other):
        '''other ** self. At each key if other is a dict, else apply to each value in self.'''
        return self._math_op(other, lambda x, y: y ** x)

    def __pos__(self):
        '''+self. Apply to each value in self.'''
        return self.apply(operator.pos)

    def __neg__(self):
        '''-self. Apply to each value in self.'''
        return self.apply(operator.neg)

    # # # DISPLAY # # #
    def __repr__(self):
        MAXLEN = self.REPR_ITEM_MAXLEN
        orig_reprs = {k: repr(v) for k, v in self.items()}
        short_reprs = dict()
        for k, v in orig_reprs.items():
            if len(v) > MAXLEN:
                short_reprs[k] = f'<{v[:MAXLEN]}...>'.replace('\n', '  ')
            else:
                short_reprs[k] = v
        contents = [f'{k!r}: {v}' for k, v in short_reprs.items()]
        return f'{type(self).__name__}({{{", ".join(contents)}}})'

    def __str__(self):
        '''like repr but never abbreviate. Also, use str of objs.'''
        orig = {k: str(v) for k, v in self.items()}
        contents = [f'{k!r}: {v}' for k, v in orig.items()]
        content_str = ",\n".join(contents)
        return f'{type(self).__name__}({{{content_str}}})'



"""
File Purpose: MultiCalculator
"""
import os

from .multi_run_tools import canon_runs
from ..tools import (
    DictOfSimilar,
)

class MultiCalculator(DictOfSimilar):
    '''a class to handle multiple calculators.

    Example:
        calc1 = PlasmaCalculator(...)
        calc2 = PlasmaCalculator(...)
        mc = MultiCalculator({'c1':calc1, 'c2':calc2})
        mc.snap = 0             # sets calc.snap=0 for all calcs
        mc['slices', 'x'] = 7   # sets calc.slices['x'] = 7 for all calcs
        mc('Eheat')             # returns list of calc('Eheat') for all calcs
    '''
    cls_new = DictOfSimilar  # results will be DictOfSimilar instead of MultiCalculator.

    def _get_similar_attrs(self):
        '''return SIMILAR ATTRS for self.
        This is the intersection of all kw_call_options from calculators in self.
        '''
        # [EFF] if this ever becomes slow, use caching.
        options = [set(calc.kw_call_options()) for calc in self.values()]
        return set.intersection(*options)

    SIMILAR_ATTRS = property(lambda self: self._get_similar_attrs(),
        doc='''SIMILAR_ATTRS for self; operations on self will be broadcasted to these attrs,
        e.g. self.snap = 0 --> [calc.snap for calc in self.values()], if 'snap' in SIMILAR_ATTRS.''')

    @property
    def title(self):
        '''return a title for self: f'{calc1.title}|{calc2.title}|...|{calcN.title}'.'''
        return '|'.join([calc.title for calc in self.values()])

    @classmethod
    def from_canon(cls, makecalc, *, dir=os.curdir, exclude=[], singles=True, abbrv=True):
        '''return MultiCalculator from canonical runs within directory (and subdirectories).
        A run is "canonical" if in a 'run' or 'runs' parameter in a _canon.txt file.
        Paths in _canon.txt file are relative to the same directory as the _canon.txt file.

        makecalc: callable
            function to make a calculator, given abspath of run's directory.
        dir: str
            directory to search for _canon.txt files.
        exclude: str, or list of strs
            exclude any subdirectories whose name equals one of these strings or re.fullmatch one of these strings.
            E.g. exclude='*[.]io' will exclude all subdirectories whose name ends with '.io';
                exclude='parallel' will exclude all subdirectories whose name equals 'parallel'.
        singles: bool
            if True, assert all run lists from _canon.txt files have length 1,
                and replace each runlist with runlist[0].
        abbrv: bool
            whether to use abbreviated names for result keys and calculator titles.
            abbreviated names determined by removing os.path.commonpath() of run dirs.
            calculator titles set via calculator.title = abbrv_name, for each calculator.
        '''
        if not singles:
            raise NotImplementedError('[TODO] from_canon when singles!=True')
        canons = canon_runs(dir=dir, exclude=exclude, singles=singles)
        result = {k: makecalc(os.path.join(k, v)) for k, v in canons.items()}
        if abbrv:
            commonpath = os.path.commonpath(canons)
            result = {os.path.relpath(k, commonpath): v for k, v in result.items()}
            for k, v in result.items():
                v.title = k
        return cls(result)



"""
File Purpose: timers, runtime profiling
"""

import cProfile
import pstats
import time
import signal

from ..errors import TimeoutError
from ..defaults import DEFAULTS


''' --------------------- Runtime Profiling --------------------- '''

class Profile(cProfile.Profile):
    '''runtime profiler. Usage:
    p = Profile()
    with p:
        >> code to profile goes here <<
    p.print_stats()  # or p.print_stats(key), key='time', 'cumulative', or other key from self.SortKey

    This whole class is a "thin" wrapper around cProfile & pstats
    For a more lightweight form of this class, you can use:
        import cProfile
        p = cProfile.Profile()
        with p:
            >> code to profile goes here <<
        p.print_stats('time')

    sortkey: 'time', 'cumulative', or other key from self.SortKey
        tells default way to sort stats during print_stats.
    clearing: bool, default True
        whether to self.clear() when starting to profile.
        (i.e., when self.enable() is called. Note that self.__enter__() calls self.enable().)
    '''
    def __init__(self, *args, sortkey='time', clearing=True, **kw):
        '''does super().__init__ but also sets self.sortkey.'''
        super().__init__(*args, **kw)
        self.sortkey = sortkey
        self.clearing = clearing

    SortKey = pstats.SortKey  # << for easy reference. See e.g. help(self.SortKey) for details.

    def print_stats(self, sortkey=None):
        '''super().print_stats() using sortkey or self.sortkey if None'''
        if sortkey is None: sortkey = self.sortkey
        super().print_stats(sortkey)

    def enable(self, *args__super, clearing=None, **kw__super):
        '''super().enable(), but first maybe call self.clear().
        clearing: None or bool, default None
            whether to first call self.clear().
            None --> use clearing = self.clearing.
        '''
        if clearing is None: clearing = self.clearing
        if clearing:
            self.clear()
        super().enable(*args__super, **kw__super)

PROFILE = Profile()
def profiling(sortkey=None):
    '''returns default profiler, PROFILE. Sets its sortkey if provided.
    see help(PROFILE.SortKey) for sortkey options.

    Example Usage:
        with profiling():
            >> code to profile goes here <<
        print_profile()
    '''
    if sortkey is not None: PROFILE.sortkey = sortkey
    return PROFILE

def print_profile(sortkey=None):
    '''prints stats from the default profiler (PROFILE).
    see help(PROFILE.SortKey) for sortkey options.

    Example Usage:
        with profiling():
            >> code to profile goes here <<
        print_profile()
    '''
    PROFILE.print_stats(sortkey)

def start_profiling():
    '''equivalent to PROFILE.enable(). Recommended: use profiling() as a context manager instead.
    Example Usage:
        start_profiling()
        >> code to profile goes here <<
        stop_profiling()
    '''
    PROFILE.enable()

def stop_profiling():
    '''equivalent to PROFILE.disable(). Recommended: use profiling() as a context manager instead.
    Example Usage:
        start_profiling()
        >> code to profile goes here <<
        stop_profiling()
    '''
    PROFILE.disable()


''' --------------------- Timers --------------------- '''

class Stopwatch():
    '''tracks time since last clear (via self.reset()), or last "marked" time (via self.mark_time()).
    All times are in seconds.

    printing: bool, default True
        whether to print self when exiting context, if self was used as a context manager.

    Can also be used as a context manager, to time a block of code:
        with Stopwatch() as watch:
            >> code to time goes here <<
        print(watch)  # or, if watch.printing=True, this would already be printed, automatically.

    Note that re-using an instance of Stopwatch as a context manager will reset it:
        with Stopwatch() as watch:
            >> code block 1 <<
        watch.time   # time it took to run code block 1
        with watch:  # <-- calls watch.reset() when entering block.
            >> code block 2 <<
        watch.time   # time it took to run code block 2
    '''
    # _t_0: time at "t=0". Set by self.reset()
    # _t_marks: "marked" times (a dict). Default has key=None.
    #             Set by self.mark_time(). Note: self.reset() also removes ALL marked times.
    # _t_create: time when self was created (never changes).
    def __init__(self, printing=True):
        self.reset()
        self._t_create = self._t_0
        self.printing = printing

    # time coordinate setters #
    def reset(self, *, _time=None):
        '''sets "t=0" time to now (or _time, if provided). Also removes all "marked" times.'''
        self._t_0 = self.now() if _time is None else _time
        self._t_marks = dict()  # << reset "marked" times.

    def mark_time(self, key=None, *, _time=None):
        '''sets "marked" time to now (or _time, if provided).
        key: hashable object, default None
            associate marked time with this key. (None --> "default" marked time.)
        '''
        self._t_marks[key] = self.now() if _time is None else _time

    # time getters #
    time = property(lambda self: self.time_since_reset, doc='''alias to self.time_since_reset''')
    time_elapsed = time   # << another alias to time_since_reset

    def time_since_reset(self):
        '''returns time [in seconds] since the "t=0" time on this watch.
        "t=0" set by self.reset(). Also set when created.
        '''
        return self.now() - self._t_0

    def time_since_mark(self, key=None):
        '''returns time [in seconds] since the "mark" time on this watch.
        "mark" set by self.mark_time(). Also set when created or reset().

        key: hashable object, default None
            "marked time" associated with this key.
        '''
        return self.now() - self.get_marked_time(key=key)

    def get_marked_time(self, key=None):
        '''returns marked time associated with this key.'''
        return self._t_marks.get(key, self._t_0)

    def get_marked_times(self):
        '''returns dict of marked times.'''
        return self._t_marks

    @staticmethod
    def now():
        '''returns time.time(). Provided as a static method for convenience.'''
        return time.time()

    # context manager #
    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.printing:
            print(self)

    # display #
    def __repr__(self):
        return f'{type(self).__name__}(time_since_reset = {self.time_since_reset():.2f} seconds)'

    def time_elapsed_string(self, as_string=False):
        '''returns string 'Total time elapsed: {:.2f} seconds.'.format(self.time_since_reset())'''
        return f'Total time elapsed: {self.time_since_reset():.2f} seconds.'


class TickingWatch(Stopwatch):
    '''Stopwatch that also "ticks" every N seconds.
    No active background process; just checks time, when queried. See self.tick().
    
    self.tick() returns whether at least N seconds have passed since the "marked" time,
        AND if the result is True, sets the "marked" time to now.
    self.nticks tracks the number of times self.tick() ever returned True.
        note: self.nticks resets to 0 when self.reset() is called.

    interval: int (possibly negative or 0)
        N. Number of seconds between ticks.
        >0 --> minimum time between calls to self.tick() returning True.
        =0  --> self.tick() will always return True.
        <0 --> self.tick() will always return False.
    wait: bool, default True
        whether the first True tick() may require waiting.
        False --> the first call to self.tick() will always give True.
        True --> the first call to self.tick() will only give True if self.interval seconds have passed.
    _t_mark_key: hashable object, default 'tick'
        key associated with the "marked" time for ticking.
        note: if None, ticks are stored in the "default marked time" key; see Stopwatch for details.
    '''
    def __init__(self, interval, *, wait=True, _t_mark_key='tick'):
        super().__init__()
        self.interval = interval
        self.wait = wait
        self.waiting = wait  # << whether we are waiting for tick.
        self._t_mark_key = _t_mark_key

    def reset(self, **kw):
        '''resets the "t=0" time, also set self.nticks=0.'''
        super().reset(**kw)
        self.nticks = 0

    def time_since_tick(self):
        '''get time since last tick. Note: does not edit the "marked" time.'''
        return self.time_since_mark(key=self._t_mark_key)

    def _mark_tick_time(self, *, _time=None):
        '''mark time of this tick. Intended for internal use, only.'''
        self.mark_time(key=self._t_mark_key, _time=_time)

    def tick(self):
        '''returns True if enough time has passed for the next tick.
        If True, also "marks" the current time, as the previous tick time.
        "Enough time" usually means at least self.interval seconds.
        Exceptions:
            if not self.waiting, return True (and set self.waiting = False)
            if interval < 0, return False.
        '''
        if self.interval < 0:
            result = False
        elif not self.waiting:
            result = True
            self.waiting = True
        else:
            result = (self.time_since_tick() >= self.interval)
        if result:
            self.nticks += 1
            self._mark_tick_time()
        return result


''' --------------------- Progress updates --------------------- '''

class ProgressUpdater(TickingWatch):
    '''class for printing messages but only when enough time has passed.
    "Enough" is defined by the input parameter print_freq (in seconds).

    print_freq: None or int (possibly negative or 0)
        Number of seconds between progress updates.
        None --> use DEFAULTS.PROGRESS_UPDATES_PRINT_FREQ
        >0 --> minimum time between calls to self.tick() returning True.
        =0  --> self.tick() will always return True.
        <0 --> self.tick() will always return False.
    wait: bool, default False
        whether to wait until print_freq seconds has passed before doing the first printout.
    print_time: bool, default False
        whether to print time by default
    clearline: bool, default True
        whether to clear the current line before printing.
    clearN: int, default 100
        number of characters to clear if clearing a line.

    Example:
        updater = ProgressUpdater(print_freq=2, wait=True)
        updater.print('This will not be printed')   # not printed because 2 seconds have not passed yet.
        time.sleep(2.5)  # << wait 2.5 seconds. (Or, put code here which takes >= 2 seconds)
        updater.print('This WILL be printed!')      # prints, then timer is reset,
        updater.print('This will not be printed')   # so it won't print again until 2 more seconds have passed.
    '''
    def __init__(self, print_freq=None, *, wait=False, print_time=True, clearline=True, clearN=100,):
        if print_freq is None:
            print_freq = DEFAULTS.PROGRESS_UPDATES_PRINT_FREQ
        super().__init__(interval=print_freq, wait=wait)
        self.print_time = print_time
        self.clearline  = clearline
        self.clearN     = clearN
    
    print_freq = property(lambda self: getattr(self, 'interval'),
                          lambda self, value: setattr(self, 'interval', value),
                          doc='''alias for self.interval''')

    def print_clear(self, N=None, force=False):
        r'''clears current printed line, and moves cursor to beginning of the line.
        only if self.clearline AND self has printed at least 1 message. OR if force.

        troubleshooting: did you use end='', e.g. print(..., end='')?
            (If not, your print statement may go to the next line.)

        N: None or int
            number of characters to clear.
            None --> use self.clearN instead.
        force: bool, default False
            whether to force print, even if not self.clearline.
            False --> require self.clearline, else don't print.

        Equivalent to:
            print('\r'+ ' '*N +'\r', end='')
        '''
        if force or (self.clearline and self.nticks > 0):
            if N is None: N = self.clearN
            print('\r'+ ' '*N +'\r', end='')

    def message_to_print(self, *args_message, sep=' '):
        '''returns message that would be printed, including info about time elapsed.
        Equivalent to message printed by self.print(...), if it's time to print an update.
        '''
        return sep.join((*args_message, self.time_elapsed_string()))

    def force_print(self, *args_message, end='', print_time=None, **kw__print):
        '''prints, without first checking whether to print.
        Clear the line first, if self.clearline.
        print_time: None or bool
            whether to print f'Total time elapsed: {format(t):.2f} seconds.'
            None --> use self.print_time (default: True)
        '''
        self.print_clear()
        if print_time or ((print_time is None) and self.print_time):
            args_message = (*args_message, self.time_elapsed_string())
        print(*args_message, end=end, **kw__print)

    def print(self, *args_message, end='', print_time=None, **kw__print):
        '''prints message given by *args_message, if it is time to print.
        Clear the line first, if self.clearline.
        print_time: None or bool
            whether to print f'Total time elapsed: {format(t):.2f} seconds.'
            None --> use self.print_time (default: True)
        '''
        if self.tick():
            self.force_print(*args_message, end=end, print_time=print_time, **kw__print)

    def printf(self, f=lambda: '', *, end='', print_time=None, **kw__print):
        '''prints message given by calling f() (which must return a single string).
        [EFF] note: only calls f() if actually printing!
            Can use this instead of self.print() if computing the print message might be expensive.

        Clear the line first, if self.clearline.
        print_time: None or bool
            whether to print f'Total time elapsed: {format(t):.2f} seconds.'
            None --> use self.print_time (default: True)

        Example:
            updater = ProgressUpdater(print_freq=2)
            updater.printf(lambda: '{} '.format(object_whose_string_is_expensive_to_calculate))
        '''
        if self.tick():
            message = f()
            self.force_print(message, end=end, print_time=print_time, **kw__print)

    def finalize(self, process_name=None, *, always=False, **kw__print):
        '''prints 'Completed process in 0.00 seconds!', filling in the process name and time elapsed as appropriate.
        Also, clear the line first, if self.clearline.

        process_name: None (default) or string
            None --> don't print any info about the process which completed.
            string --> include this name in the finalize message.
        always: False (default) or True
            False --> ignore this kwarg.
            True --> always print the finalize message.
        '''
        time_elapsed = self.time()
        if (always) or (self.nticks > 0):
            process_name = '' if process_name is None else repr(process_name) + ' '
            message = f'Completed {process_name}in {time_elapsed:.2f} seconds!'
            self.force_print(message, print_time=False, **kw__print)


''' --------------------- Execution time limiter --------------------- '''

class TimeLimit():
    '''context manager which imposes a time limit, exiting with TimeoutError if internal code takes too long.
    Usage example:
        with TimeLimit(10):
            # any code in this block must complete within 10 seconds, else will crash with TimeoutError.

    seconds: None or int
        timeout after this many seconds (unless code completes before then).
        Must be an integer, due to limitations of signal.alarm method.
        None, or <=0 --> no time limit
    '''
    def __init__(self, seconds):
        if seconds is None:
            seconds = 0
        self.seconds = seconds

    now = staticmethod(time.time)  # << for easy reference.

    def time_elapsed(self):
        '''returns time elapsed since entering the context.'''
        return self.now() - self.start_time

    def signal_handler(self, _signum, _frame):
        '''raise TimeoutError.'''
        raise TimeoutError(f"Timed out after {self.time_elapsed():.2g} seconds.")

    def __enter__(self):
        self.start_time = time.time()
        if self.seconds > 0:
            signal.signal(signal.SIGALRM, self.signal_handler)
            signal.alarm(self.seconds)  # raise signal.SIGALRM in this many seconds from now.
            self.set_alarm = True
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if getattr(self, 'set_alarm', False):
            signal.alarm(0)  # cancel any remaining alarms.



"""
File Purpose: misc tricks related to plotting

stuff that's sometimes easy to forget, but super useful if you remember it.
"""
import math

import matplotlib.pyplot as plt
import matplotlib.ticker as mpl_ticker

from ...errors import InputError


def ax_outline(ax=None, *, color='black', linewidth=1):
    '''draw a box around the axes'''
    if ax is None:
        ax = plt.gca()
    ax.patch.set(linewidth=linewidth, edgecolor=color)

def fig_outline(fig=None, *, color='red', linewidth=1):
    '''draw a box around the figure'''
    if fig is None:
        fig = plt.gcf()
    fig.patch.set(linewidth=linewidth, edgecolor=color)

def ax_remove_ticks(ax=None):
    '''remove ticks from the axes!'''
    if ax is None:
        ax = plt.gca()
    ax.set(xticks=[], yticks=[])

def ax_remove_ticklabels(ax=None):
    '''remove ticklabels from the axes (but don't remove ticks)!'''
    if ax is None:
        ax = plt.gca()
    ax.set(xticklabels=[], yticklabels=[])

def get_colorbar_axes(fig=None):
    '''return list of all colorbar axes on the figure (current figure if fig is None).
    (in the same order in which they appear in fig.axes)
    '''
    if fig is None:
        fig = plt.gcf()
    return [ax for ax in fig.axes if ax.get_label()=='<colorbar>']

# # number of ticks, and tick locations # #
def get_min_n_ticks(ax=None):
    '''get the minimum number of ticks on this axes (current axes if None provided).
    This "minimum" is according to the x and y tick locators.
    Returns (min for x-axis, min for y-axis).
    None if cannot find tick locator with a minimum number of ticks.
    '''
    if ax is None:
        ax = plt.gca()
    ticker_x = ax.xaxis.get_major_locator()
    if hasattr(ticker_x, 'base'):  # e.g., RadialLocator or ThetaLocator
        ticker_x = ticker_x.base
    result_x = getattr(ticker_x, '_min_n_ticks', None)
    ticker_y = ax.yaxis.get_major_locator()
    if hasattr(ticker_y, 'base'):
        ticker_y = ticker_y.base
    result_y = getattr(ticker_y, '_min_n_ticks', None)
    return (result_x, result_y)

def set_min_n_ticks(n, ax=None, *, fail_ok=True):
    '''set the minimum number of ticks on this axes (current axes if None provided).
    Alters the tick locators for x and y axes here, via locator.set_params(min_n_ticks=n)
    If tick locators have 'base', use locator.base.set_params(min_n_ticks=n) instead.
        (e.g. RadialLocator and ThetaLocator have 'base'.)

    n: int or 2-tuple of None or int
        minimum number of ticks to have on the x, y axes.
        single int --> equivalent to providing (n, n).
        None in tuple --> don't set min_n_ticks for that axis.
    fail_ok: bool
        whether it is okay to fail silently if not able to set min_n_ticks in the existing locators.

    returns (whether succeeded for x-axis, whether succeeded for y-axis)
        (None if None in original `n` input for that axis)

    Examples:
        # set x and y axes to each have at least 5 ticks
        set_min_n_ticks(5)

        # set y-axis to have at least 3 ticks; x-axis unaffected
        set_min_n_ticks((None, 3))

        # set all colorbars to have at least 7 ticks
        for cax in pc.get_colorbar_axes():
            pc.set_min_n_ticks(7, cax)
    '''
    if ax is None:
        ax = plt.gca()
    try:
        iter(n)
    except TypeError:
        n = (n,n)
    nx, ny = n
    # x axis
    if nx is None:
        success_x = None
    else:
        ticker_x = ax.xaxis.get_major_locator()
        if hasattr(ticker_x, 'base'):  # e.g., RadialLocator or ThetaLocator
            ticker_x = ticker_x.base
        try:
            ticker_x.set_params(min_n_ticks=nx)
        except Exception:  # not BaseException! e.g., don't catch KeyboardInterrupt.
            success_x = False
            if not fail_ok:
                raise
        else:
            success_x = True
    # y axis
    if ny is None:
        success_y = None
    else:
        ticker_y = ax.yaxis.get_major_locator()
        if hasattr(ticker_y, 'base'):
            ticker_y = ticker_y.base
        try:
            ticker_y.set_params(min_n_ticks=ny)
        except Exception:
            success_y = False
            if not fail_ok:
                raise
        else:
            success_y = True
    return (success_x, success_y)


def use_simple_log_tick_locator(ax=None, x='both', base=10):
    '''tell log-scaled axes of ax to use SimpleLogTickLocator.
    x: 'both', 'x', or 'y'.
        tells whether to apply to both (if both are currently log-scaled),
        or only x or y. If x or y but not currently log-scaled, crash.
    '''
    if ax is None:
        ax = plt.gca()
    do = []
    if x == 'both':
        if ax.get_xscale() == 'log':
            do.append('x')
        if ax.get_yscale() == 'log':
            do.append('y')
    elif x == 'x':
        if ax.get_xscale() == 'log':
            do.append('x')
        else:
            raise InputError('x-axis not log-scaled!')
    elif x == 'y':
        if ax.get_yscale() == 'log':
            do.append('y')
        else:
            raise InputError('y-axis not log-scaled!')
    else:
        raise InputError(f"Invalid x value: {x!r}. Expected 'both', 'x', or 'y'.")
    if 'x' in do:
        ax.xaxis.set_major_locator(SimpleLogTickLocator(base=base))
    if 'y' in do:
        ax.yaxis.set_major_locator(SimpleLogTickLocator(base=base))

class SimpleLogTickLocator(mpl_ticker.LogLocator):
    '''simple LogLocator: puts base^N for every next power of base.
    Includes all powers, doesn't check if it is too dense to look nice.
    Also doesn't check if it's too sparse to look nice.
    
    base: log base. default 10.
    '''
    def __init__(self, base=10):
        super().__init__(base=base)

    def tick_values(self, vmin, vmax):
        '''return all ticks from floor(log(vmin)) to ceil(log(vmax))
        (tick min and tick max won't show up on the plot unless between view lims.)
        '''
        logmin = math.floor(math.log(vmin, self._base))
        logmax = math.ceil(math.log(vmax, self._base))
        return [self._base**i for i in range(logmin, logmax+1)]


def use_simple_ticks_renamer(x, f, ax=None):
    '''simple tick formatter, changes tick names from val to f(val).
    f: callable
        f(tick value) --> string to use for labeling tick.
    x: 'x', 'y', or both
        which axis to set the tick formatter on.
    '''
    def _f_with_pos_arg(val, pos):
        return f(val)
    if ax is None:
        ax = plt.gca()
    if x == 'x':
        ax.xaxis.set_major_formatter(mpl_ticker.FuncFormatter(_f_with_pos_arg))
    elif x == 'y':
        ax.yaxis.set_major_formatter(mpl_ticker.FuncFormatter(_f_with_pos_arg))
    elif x == 'both':
        ax.xaxis.set_major_formatter(mpl_ticker.FuncFormatter(_f_with_pos_arg))
        ax.yaxis.set_major_formatter(mpl_ticker.FuncFormatter(_f_with_pos_arg))
    else:
        raise InputError(f"Invalid x value: {x!r}. Expected 'both', 'x', or 'y'.")

# # size of a drawn artist (e.g., textbox, axes...) # #
# artist.get_window_extent()

# # actual aspect ratio of an axes # #
def ax_aspect(ax):
    # Total figure size
    figW, figH = ax.get_figure().get_size_inches()
    # Axis size on figure
    _, _, w, h = ax.get_position().bounds
    # Ratio of display units
    disp_ratio = (figH * h) / (figW * w)
    # Ratio of data units
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    data_ratio = (ylim[1] - ylim[0]) / (xlim[1] - xlim[0])
    return disp_ratio / data_ratio

# # transforms (for 'transform' kwarg of plotting functions) # #
def plt_transformer(xy=('data', 'data'), *, ax=None):
    '''simple result useful for 'transform' kwarg for plotting functions.
    xy: 'data', 'axes', or 2-tuple of ('data' or 'axes')
        tells whether 'x' and 'y' should be in 'data' or 'axes' coordinates,
            when using result of `plt_transformer` as 'transform' kwarg in plotting function.
        single string --> both in this coordinate system.
        tuple --> xy[0] tells x system; xy[1] tells y system.

    'data' coords means input values match data values.
    'axes' coords means input values correspond to distance across axis:
        for x: left=0, right=1.
        for y: bottom=0, top=1.
    ax: None or matplotlib axes object
        axes to use for the transformer. If None, use plt.gca() instead.
    '''
    if ax is None:
        ax = plt.gca()
    if isinstance(xy, str):
        xy = (xy, xy)
    x, y = xy
    if x == 'data' and y == 'data':
        return ax.transData
    elif x == 'axes' and y == 'axes':
        return ax.transAxes
    elif x == 'data' and y == 'axes':
        return ax.get_xaxis_transform()
    elif x == 'axes' and y == 'data':
        return ax.get_yaxis_transform()
    else:
        raise InputError(f"Invalid x and/or y: {x!r}, {y!r}. Expected each to be 'data' or 'axes'.")

# # convenient locations to have for reference: # # 
PLOT_LOCATION_NAMES = \
    ''''outer upper left',    'above upper left', 'above upper center', 'above upper right', 'outer upper right',
    'outside upper left',  'upper left',       'upper center',       'upper right',       'outside upper right',
    'outside center left', 'center left',      'center',             'center right',      'outside center right',
    'outside lower left',  'lower left',       'lower center',       'lower right',       'outside lower right',
    'outer lower left',    'below lower left', 'below lower center', 'below lower right', 'outer lower right'.'''

def plot_locations(margin=0.03):
    '''dict of xy=(x,y) in axes coordinates (0=left/bottom, 1=right/top) for various locations;
    also includes nice values for ha & va to use for aligning text.

    margin: number, probably between 0 and 0.25
        margin to add to the plot locations. E.g. use 0+margin for bottom instead of 0.

    The locations are as follows, with outer/above/outside/below referring to locations outside the axes.
    'outer upper left',    'above upper left', 'above upper center', 'above upper right', 'outer upper right',
    'outside upper left',  'upper left',       'upper center',       'upper right',       'outside upper right',
    'outside center left', 'center left',      'center',             'center right',      'outside center right',
    'outside lower left',  'lower left',       'lower center',       'lower right',       'outside lower right',
    'outer lower left',    'below lower left', 'below lower center', 'below lower right', 'outer lower right'.
    '''
    result = {
        # inside the axes:
        'upper left':   dict(xy=(0+margin, 1-margin), ha='left',   va='top'),
        'upper center': dict(xy=(0.5     , 1-margin), ha='center', va='top'),
        'upper right':  dict(xy=(1-margin, 1-margin), ha='right',  va='top'),
        'center left':  dict(xy=(0+margin, 0.5     ), ha='left',   va='center'),
        'center':       dict(xy=(0.5     , 0.5     ), ha='center', va='center'),
        'center right': dict(xy=(1-margin, 0.5     ), ha='right',  va='center'),
        'lower left':   dict(xy=(0+margin, 0+margin), ha='left',   va='bottom'),
        'lower center': dict(xy=(0.5     , 0+margin), ha='center', va='bottom'),
        'lower right':  dict(xy=(1-margin, 0+margin), ha='right',  va='bottom'),
        # outside the axes - top row:
        'outer upper left':     dict(xy=(0-margin, 1+margin), ha='right',  va='bottom'),
        'above upper left':     dict(xy=(0+margin, 1+margin), ha='left',   va='bottom'),
        'above upper center':   dict(xy=(0.5     , 1+margin), ha='center', va='bottom'),
        'above upper right':    dict(xy=(1-margin, 1+margin), ha='right',  va='bottom'),
        'outer upper right':    dict(xy=(1+margin, 1+margin), ha='left',   va='bottom'),
        # outside the axes - left & right columns
        'outside upper left':   dict(xy=(0-margin, 1-margin), ha='right',  va='top'),
        'outside upper right':  dict(xy=(1+margin, 1-margin), ha='left',   va='top'),
        'outside center left':  dict(xy=(0-margin, 0.5     ), ha='right',  va='center'),
        'outside center right': dict(xy=(1+margin, 0.5     ), ha='left',   va='center'),
        'outside lower left':   dict(xy=(0-margin, 0+margin), ha='right',  va='bottom'),
        'outside lower right':  dict(xy=(1+margin, 0+margin), ha='left',   va='bottom'),
        # outside the axes - bottom row:
        'outer lower left':     dict(xy=(0-margin, 0-margin), ha='right',  va='top'),
        'below lower left':     dict(xy=(0+margin, 0-margin), ha='left',   va='top'),
        'below lower center':   dict(xy=(0.5     , 0-margin), ha='center', va='top'),
        'below lower right':    dict(xy=(1-margin, 0-margin), ha='right',  va='top'),
        'outer lower right':    dict(xy=(1+margin, 0-margin), ha='left',   va='top'),
        }
    return result

def plot_note(note, loc='upper left', *, margin=0.03, ax=None, xy=None, **kw_annotate):
    '''add this note to the current plot at the indicated position.
    For more detailed control, use plt.annotate instead.

    loc: str or tuple
        position in axes coords.
        str -> also sets horizontal & vertical alignment (via 'ha' & 'va' kwargs).
        See below for valid strings.
    margin: number, probably between 0 and 0.25
        margin to add to the plot locations. E.g. use 0+margin for bottom instead of 0.
    ax: None or matplotlib axes object
        axes to add the note to. If None, use plt.annotate() instead of ax.annotate().
    xy: None or alias for loc
        if provided, use xy input instead of `loc`, and completely ignore `loc`.

    Valid strings are (with outer/above/outside/below being outside the axes):
    'outer upper left',    'above upper left', 'above upper center', 'above upper right', 'outer upper right',
    'outside upper left',  'upper left',       'upper center',       'upper right',       'outside upper right',
    'outside center left', 'center left',      'center',             'center right',      'outside center right',
    'outside lower left',  'lower left',       'lower center',       'lower right',       'outside lower right',
    'outer lower left',    'below lower left', 'below lower center', 'below lower right', 'outer lower right'.
    '''
    if xy is None:
        xy = loc
    defaults = plot_locations(margin=margin)
    if isinstance(xy, str):
        if xy not in defaults:
            raise ValueError(f"Invalid xy string: {xy!r}. Expected one of:\n    {PLOT_LOCATION_NAMES}")
        kw_annotate = {**defaults[xy], **kw_annotate}
    else:
        kw_annotate['xy'] = xy
    kw_annotate.setdefault('xycoords', 'axes fraction')
    if ax is None:
        return plt.annotate(note, **kw_annotate)
    else:
        return ax.annotate(note, **kw_annotate)



"""
File Purpose: CopapicDirectLoader
"""
import os
import math  # for lcm...

# required external modules
import numpy as np
import xarray as xr

# optional external modules
from ...tools import ImportFailed
try:
    import openpmd_api as io
except ImportError as err:
    io = ImportFailed("openpmd_api", err=err, locals=locals())

# internal modules
from ...dimensions import INPUT_SNAP
from ...errors import (
    LoadingNotImplementedError,
    InputError,
)
from ...quantities import DirectLoader, QuasineutralLoader
from ...tools import interprets_fractional_indexing
from ...defaults import DEFAULTS


''' --------------------- CopapicDirectLoader --------------------- '''

class CopapicDirectLoader(DirectLoader, QuasineutralLoader):
    '''manages loading data directly from Copapic output files.

    input_deck: CopapicInputDeck
        input deck. E.g. CopapicInputDeck.from_file('copapic.json').
        Currently, requires that input_deck.filename is not None.

    attributes of self (not available at __init__)
    '''
    # parent class ordering notes: no consistent mro() if QuasineutralLoader before DirectLoader.
    #  (related to DirectLoader being a SnapHaver, but not exactly sure why the error happens.
    #   but DirectLoader first here is fine; it doesn't overlap with QuasineutralLoader methods.)

    _slice_maindims_in_load_direct = True   # [EFF] slice directly when reading h5 files, if using self.slices

    def __init__(self, input_deck, **kw_super):
        self.input_deck = input_deck
        super().__init__(**kw_super)

    # # # READ MODE AND SNAPDIR # # #
    @property
    def snapdir(self):
        '''directory containing the snapshot files.
        '''
        return self.input_deck.output_dir

    # # # LOAD FROMFILE # # #
    def _var_for_load_fromfile(self, varname_internal):
        '''return var, suitably adjusted to pass into load_fromfile().
        Here:
            append self.component if self._loading_component
            append self.fluid.N if self._loading_fluid
        E.g. 'flux' turns into 'fluxz2' when loading 'z' component and fluid 2.
        '''
        result = varname_internal
        if getattr(self, '_loading_component', False):
            result = result + str(self.component)
        if getattr(self, '_loading_fluid', False):
            result = result + str(self.fluid.N)
        return result

    def load_fromfile(self, fromfile_var, *args, snap=None, **kw):
        '''return numpy array of fromfile_var, loaded directly from file.
        use self.full_read_mode to determine which file(s) / how to read them.

        fromfile_var: str
            the name of the variable to read, adjusted appropriately for loading fromfile.
            E.g., use 'fluxz2', not 'flux', to get flux for fluid 2 and component z.
            See also: self._var_for_load_fromfile().
        snap: None, str, int, or Snap
            the snapshot number to load. if None, use self.snap.

        Example:
            fromfile_var='fluxx1', snap=7
        '''
        snap = self._as_single_snap(snap)  # expect single snap when loading from file.
        if hasattr(snap, 'exists_for') and not snap.exists_for(self):
            result = xr.DataArray(self.snap_dim.NAN, attrs=dict(units='raw'))
            return self.assign_snap_coord(result)  # [TODO][EFF] assign coords when making array instead...
        if snap is INPUT_SNAP:
            raise LoadingNotImplementedError(f'var={fromfile_var!r}, when snap is INPUT_SNAP')
        return self._openpmd_load_fromfile(fromfile_var, *args, snap=snap, **kw)

    def directly_loadable_vars(self, snap=None):
        '''return tuple of directly loadable variables.
        These are the variables that can be loaded directly from a file,
        using the current full_read_mode.

        snap: None, str, int, or Snap
            the snapshot number to load. if None, use self.snap.
        '''
        snap = self._as_single_snap(snap)  # expect single snap when loading from file.
        if hasattr(snap, 'exists_for') and not snap.exists_for(self):
            return np.nan
        result = list(self._openpmd_directly_loadable_vars(snap=snap))
        return result

    # def wheref_loadable(self, var=None, snaps=None):
    #     '''returns snaps from self.snaps where 'n', 'flux', 'nvsqr', 'vdist', 'phi' values can be loaded,
    #     for all fluids currently in self.fluid (for non-phi vars).
    #     [EFF] method is IMPLICIT, based on var_out_subcycling details for each fluid.
    #         Does not actually check the files directly, because it's expensive to do that.

    #     [TODO] account for subsampling_info too.

    #     var: None, list of str, str from ('n', 'flux', 'nvsqr', 'vdist', 'phi'), or 'all'
    #         None --> return dict of answer for each of those vars
    #         'all' --> return SnapList where all vars are loadable.
    #         list --> return SnapList where all vars in list are loadable.
    #         str --> return SnapList for this var only.
    #     snaps: None, 'current', 'current_range', 'current_n', or SnapList
    #         the snaps to consider.
    #         None --> self.snaps.
    #         'current' --> self.snap
    #         'current_range' --> self.snaps.select_between(self.snap[0].t, self.snap[-1].t)
    #         'current_n' --> 'current_range', but evently downsample result size similar to current n snap.
    #                     (note: even downsampling isn't always possible; result size won't exactly match.
    #                      useful if you want "roughly N snaps". See example below.)
    #         E.g. if self.snap [10,20,30,...,200] (len=19), and self.snaps [0,...,300] (len=301),
    #             checking where_loadable(var) which is available every 4 snaps, result will be:
    #             None --> [0,4,8,...,300] (len=76)
    #             'current' --> [20,40,...,200] (len=10)  # because 10,30,50,... aren't divisible by 4
    #             'current_range' --> [12,16,20,...,200] (len=48)
    #             'current_n' --> [12,20,28,...,200] (len=24)
    #                             # because interprets_fractional_indexing(slice(None,None,1/19), L=48)
    #                             # gives slice(0, None, 2), which is then applied to current_range result.
    #     '''
    #     result = {}
    #     downsample = False
    #     if snaps is None:
    #         snaps = self.snaps
    #     elif isinstance(snaps, str):
    #         if snaps=='current':
    #             snaps = self.snap_list()
    #         elif snaps=='current_range' or snaps=='current_n':
    #             snaplist = self.snap_list()
    #             if snaps == 'current_n': downsample = len(snaplist)
    #             snaps = self.snaps.select_between(snaplist[0].t, snaplist[-1].t)
    #         else:
    #             raise InputError(f'got str snaps={snaps!r}, expected "current" or "current_range"')
    #     combine = False
    #     KNOWN = ('n', 'flux', 'nvsqr', 'vdist', 'phi')
    #     if var is None:
    #         vars_ = KNOWN
    #     elif isinstance(var, str):
    #         if var == 'all':
    #             combine = True
    #             vars_ = KNOWN
    #         else:
    #             vars_ = [var]
    #     else:  # list of vars
    #         combine = True
    #         vars_ = var
    #     unknown = [v for v in vars_ if v not in KNOWN]
    #     if len(unknown) > 0:
    #         raise InputError(f'unsupported var(s): {unknown!r}. Expected only vars from: {KNOWN!r}')
    #     nout = self.input_deck['nout']
    #     subcycles = {}
    #     FLUID_VARS = {
    #         'n': 'part_out_subcycle',
    #         'flux': 'flux_out_subcycle',
    #         'nvsqr': 'nvsqr_out_subcycle',
    #         'vdist': 'vdist_out_subcycle',
    #     }
    #     for v, subvar in FLUID_VARS.items():
    #         if v in vars_:
    #             subcycle = 1
    #             for fluid in self.iter_fluid():
    #                 subcycle = math.lcm(subcycle, fluid.get(subvar, 1))
    #             subcycles[v] = subcycle
    #     if 'phi' in vars_:
    #         subcycles['phi'] = self.input_deck.get('phi_out_subcycle', 1)
    #     if combine:  # combine the subcycles
    #         subcycle = 1
    #         for v, sc in subcycles.items():
    #             subcycle = math.lcm(subcycle, sc)
    #         var = 'COMBINED'
    #         subcycles = {var: subcycle}
    #     for v, subcycle in subcycles.items():
    #         if subcycle == 1:
    #             result[v] = snaps
    #         else:
    #             igood = [i for i, snap in enumerate(snaps) if int(snap.s) % (subcycle * nout) == 0]
    #             result[v] = snaps[igood]
    #     # downsample maybe
    #     if downsample:
    #         for v, rr in result.items():
    #             if len(rr) > downsample:  # actually downsample this one :)
    #                 slicer = interprets_fractional_indexing(slice(None,None,1/downsample), L=len(rr))
    #                 result[v] = rr[slicer]
    #     return result[var] if isinstance(var, str) else result

    # def where_loadable(self, var=None, snaps=None):
    #     '''returns snaps from self.snaps where 'n', 'flux', 'nvsqr', 'vdist', 'phi' values can be loaded,
    #     for ALL fluids in self.fluids.
    #     Equivalent:
    #         with self.using(fluid=None):
    #             return self.wheref_loadable(var=var, snaps=snaps)

    #     [EFF] method is IMPLICIT, based on var_out_subcycling details for each fluid.
    #         Does not actually check the files directly, because it's expensive to do that.

    #     [TODO] account for subsampling_info too.

    #     var: None, list of str, str from ('n', 'flux', 'nvsqr', 'vdist', 'phi'), or 'all'
    #         None --> return dict of answer for each of those vars
    #         'all' --> return SnapList where all vars are loadable.
    #         list --> return SnapList where all vars in list are loadable.
    #         str --> return SnapList for this var only.
    #     snaps: None, 'current', 'current_range', 'current_n', or SnapList
    #         the snaps to consider.
    #         None --> self.snaps.
    #         'current' --> self.snap
    #         'current_range' --> self.snaps.select_between(self.snap[0].t, self.snap[-1].t)
    #         'current_n' --> 'current_range', but evently downsample result size similar to current n snap.
    #                     (note: even downsampling isn't always possible; result size won't exactly match.
    #                      useful if you want "roughly N snaps". See example below.)
    #         E.g. if self.snap [10,20,30,...,200] (len=19), and self.snaps [0,...,300] (len=301),
    #             checking where_loadable(var) which is available every 4 snaps, result will be:
    #             None --> [0,4,8,...,300] (len=76)
    #             'current' --> [20,40,...,200] (len=10)  # because 10,30,50,... aren't divisible by 4
    #             'current_range' --> [12,16,20,...,200] (len=48)
    #             'current_n' --> [12,20,28,...,200] (len=24)
    #                             # because interprets_fractional_indexing(slice(None,None,1/19), L=48)
    #                             # gives slice(0, None, 2), which is then applied to current_range result.
    #     '''
    #     with self.using(fluid=None):
    #         return self.wheref_loadable(var=var, snaps=snaps)

    def _openpmd_load_fromfile(self, copapic_var, *args__None, snap=None, **kw__None):
        '''return numpy array of var, loaded directly from file, using "openpmd" read_mode.

        copapic_var: str
            the name of the variable as stored in the snapshot.
        snap: None, str, int, or Snap
            the snapshot number to load.
            if None, use self.snap.

        Example:
            copapic_var='den0', snap=7
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        
        snap = self._as_single_snap(snap)
        # if copapic_var not in self._openpmd_directly_loadable_vars(snap=snap):
        #     errmsg = f'var={copapic_var!r} not recognized'
        #     raise LoadingNotImplementedError(errmsg)
        
        series = io.Series(self._openpmd_base_name(), io.Access.read_only)
        i = series.iterations[int(str(snap))]

        if '_' in copapic_var:
            split = copapic_var.split('_')
            baseName = split[0]
            component = split[1]                
        else:
            if (copapic_var[:3] == 'den'):
                baseName = copapic_var[:3]+"_"+copapic_var[3:]
                component = " Scalar"
            else:
                baseName = copapic_var
                component = " Scalar"

        if (baseName not in i.meshes):
            raise LoadingNotImplementedError(f'var={baseName!r} not found')
        var = i.meshes[baseName]

        for a, b in var.items():
            if (var.scalar):
                var = b
                break
            if a == component:
                var = b
                break
        
        if getattr(self, '_slice_maindims_in_load_direct', False):
            # [EFF] slice here instead of reading all data then slicing later.
            slices = self.slices
            slicer = tuple(slices.get(d, slice(None)) for d in self.maindims)
            result = var[slicer]
        else:
            result = var[:]

        series.flush()
        series.close()
        return result

    def _openpmd_directly_loadable_vars(self, *, snap=None):
        '''return tuple of directly loadable variables.'''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        series = io.Series(self._openpmd_base_name(), io.Access.read_only)
        i = series.iterations[int(str(snap))]
        result = [x.replace("_",'') for x in i.meshes]
        
        return tuple(result)

    def _openpmd_base_name(self):
        '''return basefrom which to load values'.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        n_fill = np.ceil(np.log10(self.input_deck['nt'])).astype(int)
        dir_ = self.input_deck.output_dir
        suffix = self.input_deck['file_extension']
        name = f'{dir_}/data_%0{n_fill}T.{suffix}'
        return name



"""
File Purpose: misc. tools related to xarray
"""
import numpy as np
import xarray as xr

from .xarray_accessors import pcAccessor
from ..arrays import np_dtype_object_to_str
from ...errors import InputError, InputConflictError, DimensionKeyError


''' --------------------- Copying --------------------- '''

@pcAccessor.register('copy_kw')
def xarray_copy_kw(array, dims=None, *, name=None, array_to_dataset=False):
    '''return dict of info suitable for creating a similar array or dataset.
    result includes dims, coords, and attrs (unchanged, copied).

    dims: None, str, or iterable of str
        if provided, only include these dims (and related coords) in the result.
        Useful if only interested in some of the dims,
            e.g. if array has x,y,z,t dims but only want to mimic dims and coords from x,t.
    name: None, bool, or str
        whether to also include name in result
        None --> True if array has name, else False
        True --> name = array.name.
        str --> name = name.
    array_to_dataset: bool
        if True, result will be suitable for creating a Dataset based on the input DataArray.
        (Equivalent to using dims=[], name=False, and deleting 'dims' key from result.)
    '''
    # handle array_to_dataset mode:
    if array_to_dataset:
        result = xarray_copy_kw(array, dims=[], name=False, array_to_dataset=False)
        del result['dims']
        return result
    # 'usual' usage
    result = dict()
    if isinstance(dims, str): dims = [dims]
    if dims is None: dims = array.dims
    elif any(d not in array.dims for d in dims):
        errmsg = (f'cannot copy_kw for dims={dims!r} because some dims not in array.dims={array.dims}.')
        raise DimensionKeyError(errmsg)
    result['dims'] = dims
    coords = dict()
    for cname, cval in array.coords.items():
        if len(cval.dims) == 0:  # scalar coord; always keep!
            coords[cname] = cval
        elif len(cval.dims) == 1 and cval.dims[0] in dims:  # 1D coord, and relevant
            coords[cname] = cval
        elif all(d in dims for d in cval.dims):  # 2D+ coord, and all relevant dims exist
            coords[cname] = (tuple(cval.dims), cval.data)
    result['coords'] = coords
    attrs = array.attrs.copy()
    result['attrs'] = attrs
    if name is None:
        name = hasattr(array, 'name')
    if name == True:
        name = array.name
    if name != False:
        result['name'] = name
    return result

@pcAccessor.register('with_data', totype='array')
def xarray_with_data(xarray, data, dims=None, *, name=None):
    '''return data as an xr.DataArray with dims, coords, name, and attrs copied from xarray.
    Similar to DataArray.copy(data=data), but here allows a bit more control.

    dims: None, str, or iterable of str
        if provided, only include these dims (and related coords) in the result.
        Useful if only interested in some of the dims,
            e.g. if array has x,y,z,t dims but only want to mimic dims and coords from x,t.
    name: None, bool, or str
        whether to also include name in result
        None --> True if array has name, else False
        True --> name = array.name.
        str --> name = name.
    '''
    if not isinstance(xarray, xr.DataArray):
        raise InputError(f'xarray_with_data expects args[0] to be a DataArray, not {type(xarray)}.')
    kw = xarray_copy_kw(xarray, dims=dims, name=name)
    return xr.DataArray(data, **kw)


''' --------------------- Converting Dataset to Array --------------------- '''

@pcAccessor.register('as_array')
def xarray_as_array(array):
    '''return array if DataArray, else array[var] if array is Dataset with one var.
    if array is Dataset with multiple vars, crash with InputError.
    '''
    if isinstance(array, xr.DataArray):
        return array
    elif isinstance(array, xr.Dataset):
        if len(array) == 1:
            return array[list(array)[0]]
        else:
            errmsg = ('as_array(Dataset) only possible if Dataset has exactly 1 data_var, '
                     f'but got Dataset with {len(array)} vars: {list(array.data_vars)}.')
            raise InputError(errmsg)
    else:
        raise InputError(f'cannot convert object of type {type(array)} to DataArray.')


''' --------------------- Vars from dataset --------------------- '''

@pcAccessor.register('vars_lookup', totype='dataset')
def xarray_vars_lookup(ds, lookup={}, **lookup_as_kw):
    '''return dict of {key from lookup: ds[lookup[key]]}
    result vals will be None for keys not found in ds.

    lookup: dict of {key: None, varname, or list of varnames}
        look for these varnames in ds. if list, use first varname found.
        if None, use key as varname.
    '''
    if lookup and lookup_as_kw:
        raise InputConflictError('provide lookup or lookup_as_kw, not both!')
    if lookup_as_kw:
        lookup = lookup_as_kw
    result = {}
    for key, find in lookup.items():
        result[key] = None
        if find is None:
            find = key
        if isinstance(find, str):
            find = [find]
        for varname in find:
            if varname in ds:
                result[key] = ds[varname]
                break
    return result

@pcAccessor.register('vars_lookup_with_defaults', totype='dataset')
def xarray_vars_lookup_with_defaults(ds, provided, lookup={}, defaults={}):
    '''return dict of {key from lookup: value for that key}.
    
    provided: dict of {key: None, str, or other value}
        provided values or lookup instructions for each key.
        None --> use instructions from lookup if key in lookup,
                 else value from defaults if key in defaults.
        str --> varname; result uses ds[varname]. (crash if not possible.)
        callable --> f of one arg; result uses f(ds).
        other value --> result uses this value directly.
    lookup: dict of {key: None, str, or list of str}
        'default' lookup instructions. Used when provided[key] is None.
    defaults: dict of {key: value or callable}
        default values for each key.
        Used when provided[key] is None and ds[lookup[key]] not found.
        callable --> called with ds as arg to get value.

    Examples:
        xarray_vars_lookup_with_defaults(ds, {}, {'n': None}, {'n': 100})
        # {'n': ds['n']} if ds['n'] exists, else {'n': 100}
        
        xarray_vars_lookup_with_defaults(ds, {}, {'n': 'var1'}, {'n': lambda ds: ds.size})
        # {'n': ds['var1']} if ds['var1'] exists, else {'n': ds.size}

        xarray_vars_lookup_with_defaults(ds, {'n': 'var1'}, {'n': 'var2'}, {'n': 7})
        # {'n': ds['var1']} if ds['var1'] exists, else crash.
        # (lookup & defaults are fully ignored, if non-None value in provided.)
    '''
    result = {}
    all_keys = set(provided) | set(lookup) | set(defaults)
    for key in all_keys:
        prov = provided.get(key, None)
        if prov is None:
            find = lookup.get(key, None)
            if find is None:
                find = key
            if isinstance(find, str):
                find = [find]
            for varname in find:
                if varname in ds:
                    result[key] = ds[varname]
                    break
            else:
                if key in defaults:
                    val = defaults[key]
                    if callable(val):
                        val = val(ds)
                    result[key] = val
                else:
                    errmsg = (f'no value for key={key!r}; not provided directly, '
                              f'varnames {find} not in ds, and no default provided.')
                    raise InputError(errmsg)
        elif isinstance(prov, str):
            if prov in ds:
                result[key] = ds[prov]
            else:
                errmsg = f'provided varname={prov!r} (for key={key!r}) not found in ds.'
                raise InputError(errmsg)
        elif callable(prov):
            result[key] = prov(ds)
        else:
            result[key] = prov
    return result


''' --------------------- Math Checks --------------------- '''

@pcAccessor.register('where_finite')
def xarray_where_finite(array):
    '''returns array, masked with NaNs anywhere that the values are not finite.'''
    return array.where(np.isfinite)


''' --------------------- Type Casting --------------------- '''

@pcAccessor.register('astypes', totype='dataset')
def xarray_astypes(ds, var_to_dtype):
    '''return copy of ds with var.astype(dtype) for each var, dtype in var_to_dtype.items()'''
    to_assign = {dvar: ds[dvar].astype(dtype) for dvar, dtype in var_to_dtype.items()}
    return ds.assign(**to_assign)

@pcAccessor.register('convert_types')
def xarray_convert_types(array, oldtype_to_newtype):
    '''return copy of array or dataset, using var.astype(newtype) for any var with oldtype,
    for oldtype, newtype in oldtype_to_newtype.items().
    '''
    if isinstance(array, xr.Dataset):
        ds = array
        to_convert = dict()
        for var, val in ds.items():
            for oldtype, newtype in oldtype_to_newtype.items():
                if val.dtype == oldtype:
                    to_convert[var] = newtype
                    break
        return xarray_astypes(ds, to_convert)
    else:
        for oldtype, newtype in oldtype_to_newtype.items():
            if array.dtype == oldtype:
                return array.astype(newtype)
        else:  # no match found during loop
            return array

@pcAccessor.register('dtype_object_to_str', totype='array')
def xarray_dtype_object_to_str(array):
    '''convert array with dtype=object to dtype=str, by doing str(x) for every x in array.'''
    return array.copy(data=np_dtype_object_to_str(array.values))


''' --------------------- Converting coords to strings --------------------- '''

@pcAccessor.register('object_coords_to_str')
def xarray_object_coords_to_str(array, *, maxlen=None, ndim_min=1):
    '''return copy of array with coords (of dtype=object) converted to string.
    maxlen: None or int>=5
        if int, truncate longer strings to this length-3, and add '...' to end.
    ndim_min: int
        minimum number of dimensions for a coord to be converted.
        e.g. ndim_min=1 --> coords with ndim=0 will not be altered
    '''
    to_assign = {}
    for cname, cc in array.coords.items():
        if cc.dtype == object and cc.ndim >= ndim_min:
            if cc.ndim >= 2:
                raise NotImplementedError('object_coords_to_str does not yet support ndim>=2.')
            cvals = [str(v) for v in cc.values]
            if maxlen is not None:
                cvals = [v[:maxlen-3] + '...' if len(v) > maxlen else v for v in cvals]
            to_assign[cname] = cc.copy(data=cvals)
    return array.assign_coords(to_assign)



"""
File Purpose: Bifrost Calculator
"""
import os

import numpy as np

from .bifrost_bases import BifrostBasesLoader
from .bifrost_direct_loader import BifrostDirectLoader
from .bifrost_eos_loader import BifrostEosLoader
from .bifrost_efield import BifrostEfieldLoader
from .bifrost_io_tools import (
    bifrost_infer_snapname_here,
    read_bifrost_meshfile,
    BifrostDataCutter,
)
from .bifrost_snaps import BifrostSnapList, BifrostScrSnap
from .bifrost_stagger import BifrostStaggerable
from .bifrost_units import BifrostUnitsManager
from ...errors import (
    FileAmbiguityError, FileContentsError, FileContentsMissingError, FileContentsConflictError,
    DimensionError,
)
from ...mhd import MhdCalculator
from ...tools import simple_property, UNSET

class BifrostCalculator(BifrostEfieldLoader, BifrostEosLoader, BifrostBasesLoader,
                        BifrostStaggerable,
                        BifrostDirectLoader,
                        MhdCalculator):
    '''PlasmaCalculator for Bifrost outputs.

    snapname: None or str
        snapname. Snaps info from snapname_NNN.idl files (which should have param snapname)
        None --> infer; look for files like "*_NNN.idl" in dir. if none, raise FileNotFoundError;
                if 2+ different implied snapnames, raise FileAmbiguityError.
    dir: str
        directory where files are located. Stored as self.dirname = os.path.abspath(dir).
    init_checks: bool
        whether to check some things during init.
        Use False for a small efficiency improvement if you know everything looks correct.
    units: 'si', 'cgs', or 'raw'
        units system for outputs of self. 'raw' for same as bifrost data, 'si' for SI units.
        'cgs' for cgs, but note cgs electromagnetic quantities are not yet implemented.
    '''
    def __init__(self, snapname=None, *, dir=os.curdir, init_checks=True, units='si', **kw_super):
        if snapname is None:
            snapname = bifrost_infer_snapname_here(dir)
        self.snapname = snapname
        self.dirname = os.path.abspath(dir)
        if not any(f.startswith(snapname) for f in os.listdir(self.dirname)):
            raise FileNotFoundError(f'No files found in {self.dirname} starting with {snapname!r}.')
        self.init_snaps()
        # note: self.u units depends on self.params, so it must go after init_snaps
        self.u = BifrostUnitsManager.from_bifrost_calculator(self, units=units)
        # [EFF] u defined before super() --> super() does not create new UnitsManager u.
        super().__init__(**kw_super)
        if init_checks:
            self.init_checks()

    # # # DIMENSIONS SETUP --- SNAPS # # #
    def init_snaps(self):
        '''set self.snaps based on snap files in self.dirname.
        if scr snap exists, set self.scr_snap to that snap.
        '''
        self.snaps = BifrostSnapList.from_here(self.snapname, dir=self.dirname)
        self.scr_snap = BifrostScrSnap.from_here(self.snapname, dir=self.dirname, missing_ok=True)

    # # # PARAMS # # #
    @property
    def params(self):
        '''return the global params shared by all snaps.
        Equivalent to self.snaps.params_global(recalc=False).
        '''
        return self.snaps.params_global(recalc=False)

    # # # DIMENSIONS SETUP -- MAINDIMS # # #
    @property
    def maindims(self):
        '''return tuple of maindims of self, e.g. ('x', 'y', 'z').
        if self.squeeze_direct, discards dims with size 1.
        (if set self.maindims=value, use that value instead.
            del self.maindims to restore "compute maindims corresponding to data" behavior.)
        '''
        return getattr(self, '_maindims', self._maindims_post_squeeze())
    @maindims.setter
    def maindims(self, value):
        self._maindims = value
    @maindims.deleter
    def maindims(self):
        if hasattr(self, '_maindims'):
            del self._maindims

    def get_maindims_coords(self):
        '''return dict of {'x': xcoords, 'y': ycoords, 'z': zcoords}.
        Units will be whatever is implied by self.coords_units system (default: self.units)
        coords will be sliced according to self.slices, if relevant.
        '''
        maindims = self.maindims  # e.g. ('x', 'y', 'z')
        if (self.meshfile is not None) and os.path.isfile(self.meshfile):
            result = self.load_mesh_coords()
            result = {x: result[x] for x in maindims}
        else:  # default to np.linspace.
            dx = {x: self.params[f'd{x}'] * self.u('l', self.coords_units_explicit) for x in maindims}
            Nx = {x: self.params[f'm{x}'] for x in maindims}
            result = {x: np.arange(Nx[x]) * dx[x] for x in maindims}
        result = self._apply_maindims_slices_to_dict(result)
        u_l = self.u('l', self.coords_units_explicit)
        result = {x: result[x] * u_l for x in result}
        return result

    @property
    def meshfile(self):
        '''abspath to meshfile, if self.params['meshfile'] exists, else None.'''
        meshfile = self.params.get('meshfile', None)
        if meshfile is not None:
            meshfile = os.path.join(self.dirname, meshfile)
        return meshfile

    def has_meshfile(self):
        '''tells whether self.meshfile is an existing file.'''
        meshfile = self.meshfile
        if meshfile is None:
            return False
        return os.path.isfile(meshfile)

    def load_mesh_coords(self, *, recalc=False):
        '''return dict of coords and diff (e.g. dx), from meshfile. [raw] units.
        {'x': xcoords, 'y': ycoords, 'z': zcoords, 'dx': dx, 'dy': dy, 'dz': dz}

        recalc: bool
            whether to, if possible, return cached value (not a copy of it - don't edit directly!)
        '''
        # caching
        if not recalc:
            if self.flip_z_mesh:
                if hasattr(self, '_mesh_coords_flipz'):
                    return self._mesh_coords_flipz
            else:
                if hasattr(self, '_mesh_coords_noflip'):
                    return self._mesh_coords_noflip
        # computing result
        meshfile = self.meshfile
        mesh = read_bifrost_meshfile(meshfile)
        result = {'x': mesh['x'],
                  'y': mesh['y'],
                  'z': mesh['z'],
                  'dx': mesh['x_ddup'],
                  'dy': mesh['y_ddup'],
                  'dz': mesh['z_ddup']}
        if self.flip_z_mesh:
            result['z'] = -result['z']
        # sanity checks
        if mesh['x_size'] != self.params['mx']:
            raise DimensionError(f"x meshfile size ({mesh['x_size']}) != mx param ({self.params['mx']})")
        if mesh['y_size'] != self.params['my']:
            raise DimensionError(f"y meshfile size ({mesh['y_size']}) != my param ({self.params['my']})")
        if mesh['z_size'] != self.params['mz']:
            raise DimensionError(f"z meshfile size ({mesh['z_size']}) != mz param ({self.params['mz']})")
        # caching
        if self.flip_z_mesh:
            self._mesh_coords_flipz = result
        else:
            self._mesh_coords_noflip = result
        return result

    cls_behavior_attrs.register('flip_z_mesh', default=True)
    flip_z_mesh = simple_property('_flip_z_mesh', default=True,
            doc='''whether to flip (multiply by -1) the z mesh coordinates relative to meshfile.
            When True, z<0 implies "below photosphere", for solar simulations.''')

    # # # INIT CHECKS # # #
    def init_checks(self):
        '''checks some things:
        should always be true:
            - self.snapname == snapname param value in all snaps.
            - mx, my, and mz are the same for all snapshots

        raise NotImplementedError in any of the following scenarios:
            - boundarychk enabled in any snap (or boundarychkx or boundarychky)
            - do_out_of_eq enabled in any snap
        '''
        params = self.params  # <-- checks require to get global params (from snaps)
        # # SHOULD ALWAYS BE TRUE # #
        # check snapname
        if 'snapname' not in params:
            raise FileContentsConflictError("'snapname' param is not the same for every snap.")
        if params['snapname'] != self.snapname:
            raise FileContentsError("self.snaps 'snapname' is not the same as self.snapname.")
        # check mx, my, mz
        for mx in ('mx', 'my', 'mz'):
            if mx not in params:
                raise FileContentsMissingError(f'{mx!r} param does not have the same value in every snap.')
        # # NOT YET IMPLEMENTED # #
        enabled_not_yet_implemented = ('boundarychk', 'boundarychkx', 'boundarychky',
                                        #'do_out_of_eq',
                                        )
        for p in enabled_not_yet_implemented:
            for val in self.snaps.iter_param_values(p, False):
                if val: raise NotImplementedError(f'BifrostCalculator when {p!r} enabled.')
        # # ALL CHECKS PASSED # #
        return True

    # # # INPUT TABLES (self.tabin) # # #
    @property
    def tabinputfile(self):
        '''abspath to tabinputfile (used when creating default self.tabin).
        raise FileAmbiguityError if self.params['tabinputfile'] doesn't exist.
        '''
        result = self.params.get('tabinputfile', None)
        if result is None:
            errmsg = ("cannot determine tabinputfile (required when making self._default_tabin()),\n"
                      "when 'tabinputfile' param doesn't match (& exist) across all snaps.")
            raise FileAmbiguityError(errmsg)
        result = os.path.join(self.dirname, result)
        return result

    # # # CUTTING DATA # # #
    # [TODO] maybe this should be inherited from a parent of BifrostCalculator instead...
    # [TODO] cut_data() method, like self.data_cutter().cut(), but also loop over self.snap.
    data_cutter_cls = BifrostDataCutter

    def data_cutter(self, snap=None, *, new_snapname=UNSET):
        '''return BifrostDataCutter for this snap (or current snap if None provided).
        Use result.cut() to create the cut data files.

        (cut data files are like data files but cut to the region indicated by self.slices.)

        new_snapname: str or UNSET
            new snapname to use for the cutted data files. UNSET --> '{self.snapname}_cut'
        '''
        snap = self._as_single_snap(snap)
        return self.data_cutter_cls(snap, self, new_snapname=new_snapname)



"""
File Purpose: code associated with Bifrost Snaps
"""
import os
import re

from .bifrost_io_tools import (
    read_bifrost_snap_idl,
    bifrost_snap_idl_files,
    bifrost_infer_snapname_here,
    BifrostVarPathsManager,
    BifrostScrVarPathsManager,
)
from ...dimensions import (
    ParamsSnap, ParamsSnapList,
    SpecialDimensionValue, MISSING_SNAP,
)
from ...errors import FileContentsConflictError


''' --------------------- BifrostSnap and BifrostSnapList --------------------- '''

class BifrostSnap(ParamsSnap):
    '''info about a single bifrost snapshot, including label (str) and index (int).

    The "index" should only be meaningful in the context of a SnapList
    The "label" should be the str name for this snapshot
        - unique within context (e.g. there's only one "snapshot 1" in a simulation)
        - easiest to use str int labels (e.g. "snapshot 1" --> label="1")

    s: the label (str) of the snapshot. if None, cannot convert to str.
    i: the "index" (int) of the snapshot (within a SnapList). if None, cannot convert to int.
    t: time at this snapshot ['raw' units]
    params: dict of parameters (e.g. from .idl file) at this snapshot. if None, make empty dict.
    '''
    var_paths_manager_cls = BifrostVarPathsManager

    @classmethod
    def _check_idl_filename_pattern(cls, snapname, filename):
        '''check that filename matches snapname_NNN.idl pattern. Crash if not.
        subclass might override, to avoid, enhance, or alter this check.
        returns match object from re.fullmatch.
        '''
        pattern = rf'{snapname}_([0-9]+)[.]idl'
        match = re.fullmatch(pattern, os.path.basename(filename))
        if match is None:
            errmsg = (f'Filename does not match snapname_NNN.idl pattern. '
                      f'snapname={snapname!r} filename={filename!r}')
            raise FileContentsConflictError(errmsg)
        return match

    @classmethod
    def from_idl_file(cls, filename, *, i=None):
        '''return BifrostSnap based on idl file.
        filename: str.
            file to read from. Like snapname_NNN.idl.
            will use s=NNN as label. E.g. snapname_071.idl --> s='071'.
            snapname determined from file contents (one of the params should be "snapname")
        i: None or int
            index of this snapshot in a SnapList.
        '''
        params = read_bifrost_snap_idl(filename)
        snapname = params['snapname']
        match = cls._check_idl_filename_pattern(snapname, filename)
        s = match.group(1)
        t = params.get('t', None)
        return cls(s=s, i=i, t=t, params=params)

    # # # PROPERTIES # # #
    @property
    def snapname(self):
        '''return snapname of this snap: self.params["snapname"]'''
        return self.params['snapname']

    @property
    def filename(self):
        '''return filename of idl file for this snap: snapname_NNN.idl'''
        return f'{self.snapname}_{self.s}.idl'

    # # # VAR PATH MANAGER # # #
    def var_paths_manager(self, calculator):
        '''return BifrostVarPathsManager for this snap.
        calculator: BifrostCalculator. used for snapdir.
        '''
        return self.var_paths_manager_cls(self, calculator)


class BifrostSnapList(ParamsSnapList):
    '''list of BifrostSnap objects'''
    value_type = BifrostSnap

    @classmethod
    def from_idl_files(cls, filenames):
        '''return BifrostSnapList based on idl files.
        order within result matches order of filenames.

        filenames: list of str.
            files to read from. Like snapname_NNN.idl.
            will use s=NNN as label. E.g. snapname_071.idl --> s='071'.
            snapname determined from file contents (one of the params should be "snapname")
        '''
        snaps = [cls.value_type.from_idl_file(f, i=i) for i, f in enumerate(filenames)]
        return cls(snaps)

    @classmethod
    def from_here(cls, snapname=None, *, dir=os.curdir):
        '''return BifrostSnapList from all snapname_NNN.idl files in directory.
        Sorted by snap number.

        snapname: None or str.
            snaplist from matching snapname_NNN.idl files. (NNN can be any integer, doesn't need to be 3 digits.)
            None --> infer; look for files like "*_NNN.idl" in dir;
                    if 0 such files or 2+ different implied snapnames, raise FileAmbiguityError.
        dir: str. directory to search in.
        '''
        if snapname is None: snapname = bifrost_infer_snapname_here(dir)
        filenames = bifrost_snap_idl_files(snapname, dir=dir)
        result = cls.from_idl_files(filenames)
        # quick check that snapname inside idl file 0 matches with idl file snapname.
        # (don't check all of them here because that could be somewhat slow.)
        if len(result) > 0 and result[0]['snapname'] != snapname:
            errmsg = (f'Snapname provided {snapname!r} does not match snapname param inside '
                      f'idl file {result[0].filename!r}: snapname={result[0]["snapname"]!r}')
            raise FileContentsConflictError(errmsg)
        return result

    # # # MISC PROPERTIES # # #
    @property
    def snapname(self):
        '''return snapname of this snaplist: self[0].snapname.
        returns None if no snaps in list.
        '''
        return self[0].snapname if len(self) > 0 else None


''' --------------------- Special  --------------------- '''

class BifrostScrSnap(BifrostSnap, SpecialDimensionValue):
    '''BifrostSnap corresponding to the ".scr" snapshot.
    There is at most 1 scr snap per Bifrost simulation.
    '''
    var_paths_manager_cls = BifrostScrVarPathsManager

    @classmethod
    def _check_idl_filename_pattern(cls, snapname, filename):
        '''check that filename matches snapname.idl.scr pattern. Crash if not.
        subclass might override, to avoid, enhance, or alter this check.
        '''
        pattern = rf'{snapname}.idl.(scr)'  # (scr) as group --> snap name will be 'scr'
        match = re.fullmatch(pattern, os.path.basename(filename))
        if match is None:
            errmsg = (f'Filename does not match snapname.idl.scr pattern. '
                      f'snapname={snapname!r} filename={filename!r}')
            raise FileContentsConflictError(errmsg)
        return match

    @classmethod
    def from_here(cls, snapname, *, dir=os.curdir, missing_ok=True):
        '''return BifrostScrSnap from the scr file with this snapname, in this directory.
        missing_ok: bool
            whether it is okay for scr file to be missing.
            True & missing --> return MISSING_SNAP.
            False & missing --> raise FileNotFoundError.
        '''
        scr_filename = f'{snapname}.idl.scr'
        scr_filepath = os.path.join(dir, scr_filename)
        if not os.path.exists(scr_filepath):
            if missing_ok: return MISSING_SNAP
            raise FileNotFoundError(f'Scr file not found: {scr_filepath!r}')
        return cls.from_idl_file(scr_filepath)



"""
File Purpose: Sentinel values (like None, but not built-in to Python)
"""
import sys

class Sentinel():
    '''Unique sentinel values. adapted from PEP 661.

    name: str
        str defining this sentinel. e.g. 'UNSET'
    repr: None or str
        str defining repr for this sentinel. None --> use name.
    module_name: None or str
        name of module where this sentinel was defined. None --> infer it.
    unique_includes_module: bool, default False
        whether to include module name in "uniqueness" of this sentinel.
        True --> calling Sentinel(name0) in module1 is a different object than Sentinel(name0) called in module2.
        False --> Sentinel(name0) produces the same object regardless of where it was called from.
        default False --> can access any sentinel value from any module, by providing its string.
            e.g. Sentinel('UNSET') is UNSET, always.

    will crash if name is already associated to a Sentinel which is not an instance cls. Example:
        class Subclass1(Sentinel): pass
        class Subclass2(Sentinel): pass
        A = Sentinel('SENTINEL_A')
        Subclass1('SENTINEL_A')  # crashes; A is not an instance of Subclass1.
        A is Sentinel('SENTINEL_A')  # True, and no crash.
        B = Subclass1('SENTINEL_B')
        B is Sentinel('SENTINEL_B')  # True; B is an instance of Sentinel, because Subclass1 is a subclass of Sentinel
        Subclass2('SENTINEL_B')  # crashes; B is not an instance of Subclass2.
    '''
    _registry = {}  # registry of ALL Sentinel objects. (key: module_name-name, value: sentinel)
                    # editing this dictionary directly might have unexpected consequences.

    def __new__(cls, name, repr=None, module_name=None, unique_includes_module=False):
        '''return new Sentinel object.'''
        # note, would've liked to use '*' before module_name here,
        #    but not sure how to handle keyword-only arguments in __reduce__.
        name = str(name)
        repr = str(repr) if repr else name
        if module_name is None:
            try:
                module_name = sys._getframe(1).f_globals.get('__name__', '__main__')
            except (AttributeError, ValueError):
                module_name = __name__

        if unique_includes_module:
            registry_key = f'{module_name}-{name}'
        else:
            registry_key = name

        sentinel = cls._registry.get(registry_key, None)
        if sentinel is not None:
            if isinstance(sentinel, cls):
                return sentinel
            else:
                errmsg = (f'Sentinel with name={name!r} already exists, but is not an instance of cls={cls}.'
                          f'\nThe existing sentinel is: {sentinel!r}, an object with type={type(sentinel)}.')
                raise TypeError(errmsg)

        sentinel = super().__new__(cls)
        sentinel.name = name
        sentinel.repr = repr
        sentinel.module_name = module_name
        sentinel.unique_includes_module = unique_includes_module

        return cls._registry.setdefault(registry_key, sentinel)

    def __repr__(self):
        return self.repr

    def __reduce__(self):
        return (
            self.__class__,
            (
                self.name,
                self.repr,
                self.module_name,
                self.unique_includes_module,
            ),
        )

UNSET = Sentinel('UNSET')
NO_VALUE = Sentinel('NO_VALUE')
ATTR_UNSET = Sentinel('ATTR_UNSET')
RESULT_MISSING = Sentinel('RESULT_MISSING')



"""
Package Purpose: PlasmaCalculator for MURAM.
"""
from .muram_bases import MuramBasesLoader
from .muram_calculator import MuramCalculator
from .muram_direct_loader import MuramDirectLoader
from .muram_eos_loader import MuramEosLoader
from .muram_io_tools import (
    muram_snap_files, read_muram_header, muram_directly_loadable_vars,
)
from .muram_multifluid import MuramMultifluidCalculator
from .muram_snaps import MuramSnap, MuramSnapList



"""
File Purpose: EbysusTemperaturesLoader

[TODO][MV] this probably belongs in a different file;
most of it would also apply to any other fluid models.
"""

from ...quantities import QuantityLoader

class EbysusTemperaturesLoader(QuantityLoader):
    '''temperature calculations.'''
    @known_var(deps=['Tjoule', 'u', 'm'])
    def get_vsqr(self):
        '''Average particle v^2 amongst all particles in each grid cell,
        as if the grid cell had particles in it (i.e., can be inferred even for fluid models).

        vsqr = (kB Ta / m) + u^2
        Ta = anisotropic temperatures. For Ebysus, use Ta == T.

        Inferred from temperature & velocity in cell:
            kB Ta = m * <(v - u)^2>, where u = <v>. This is the definition of Ta.
                Here, the averaging is taken across many particles.
                (T within a grid cell includes only the particles in the cell;
                "global" T across many cells includes all particles in all those cells.)
            Simplifying yields kB Ta == m * (<v^2> - 2<v>u + u^2) == m * (<v^2> - u^2).
            Thus, it is possible to infer <v^2> within a cell from T aand u in that cell:
                <v^2>_cell = (kB Ta_cell / m) + u_cell^2
        '''
        Tjoule = self('Tjoule')  # == kB * T   # uses Ta == T for Ebysus.
        return (Tjoule / self('m')) + self('u')**2

    @known_pattern(r'T(a)?(joule)?_global', deps=['vsqr', 'nmean_u', 'm'])
    def get_T_global(self, var, *, _match=None):
        '''quantity whose density-weighted mean is the global temperature,
        when all particles across the box are considered, as if the system had particles in it.

        Tjoule --> result will be in energy units.
        T (not Ta) --> take rms of components of result, e.g. sqrt((Tx^2 + Ty^2 + Tz^2) / 3)

        Tajoule_global = m * (vsqr - (nmean_u)^2)
            == kB Ta + m*u^2 - m*(nmean_u)^2
        Note that nmean(Ta_global) does NOT equal nmean(Ta). Instead:
            nmean(kB Ta_global) = nmean(kB Ta) + m*nmean(u^2) - m*(nmean_u)^2

        "DERIVATION" / EXPLANATION:
            kB Ta = m * <(v - u)^2>, where u = <v>.
            Simplifying yields kB Ta == m * (<v^2> - 2<v>u + u^2) == m * (<v^2> - u^2).
            Note that taking an average across all particles is equivalent to a
                density-weighted average across all cells, as density scales with number of particles.
                (This doesn't apply directly to T because particles don't have individual values of T,
                but it does apply to <v^2> and <v> because particles do have individual values of v.)
            Using the equation above to infer T across all particles:
                kB nmean_Ta_global = m * <v^2> - <v>^2,
                where <v^2> are the means of cell values, taken across all particles:
                --> kB nmean_Ta_global = m * (nmean(<v^2>_cell) - (nmean_u)^2)

            Ta_global is defined as the quantity whose nmean satisfies the relationship above. i.e.:
                kB Ta_global = m * <v^2>_cell - nmean_u^2

            As described in get_vsqr, we can also infer <v^2> within a cell:
                <v^2>_cell = (kB Ta_cell / m) + u_cell^2
            If desired, these can be combined, to see:
                kB Ta_global = kB Ta + m * (u_cell^2 - nmean_u^2)
        '''
        anisotropic, joule = _match.groups()
        # if getattr(self, '_debug_Ta_global', False):
        #     print('_debugging - using alternate method for Ta_global')
        #     nmean_u = self('nmean_u')
        #     u2 = self('u**2')
        #     if joule:
        #         return self('Tjoule') + self('m/kB') * (u2 - nmean_u**2)
        #     else:
        #         return self('T') + self('m/kB') * (u2 - nmean_u**2)
        # else:
        const = self('m') if joule else self('m/kB')
        result = const * (self('vsqr') - self('nmean_u')**2)
        return result if anisotropic else self.rmscomps(result)

    @known_pattern(r'T(a)?(joule)?_global_correction', deps=['u', 'mean_u', 'm'])
    def get_T_global_correction(self, var, *, _match=None):
        '''m*(u^2 - (nmean_u)^2) / kB
        Ta_global_correction + Ta == Ta_global.

        if Tajoule, don't divide by kB (answer will be in energy units).
        it T (not Ta), take rms of components.

        see self.help('T_global') for more detailed explanation.
        '''
        anisotropic, joule = _match.groups()
        const = self('m') if joule else self('m/kB')
        result = const * (self('u')**2 - self('nmean_u')**2)
        return result if anisotropic else self.rmscomps(result)

    @known_var(deps=['rmscomps_nmean_Ta_global'], ignores_dims=['component'])
    def get_T_box(self):
        '''temperature of the entire simulation box, as if full of particles,
        and observed by something that could not resolve the individual cells.

        Ignores Ta components for directions in which the box has no extent.
        (E.g. 512x512x1 box --> ignore Tz.)

        See self.help('rmscomps_nmean_Ta_global') for more details.
        '''
        return self('rmscomps_nmean_Ta_global', component=self.maindims_with_extent())



"""
Package Purpose: stuff related to single-fluid MHD.
"""

from .elements import Element, ElementList
from .mhd_bases import MhdBasesLoader
from .mhd_calculator import MhdCalculator
from .mhd_eos_loader import MhdEosLoader
from .mhd_er_tables import (
    erTable, erTableFromMemmap, erTableManager,
    eos_file_tables, rad_file_tables,
    erTabInputManager,
)
from .mhd_units import MhdUnitsManager
from .multifluid import (
    ## fluids ##
    ElementHandler, MhdFluid, ElementHaver,
    Specie, IonMixture,
    ElementHandlerList, MhdFluidList, ElementHaverList,
    SpecieList, IonMixtureList,
    SPECIES, ION_MIXTURES, ION_MIXTURE_SPECIES,
    ## genrad ##
    GenradTable, GenradTableManager,
    ## multifluid_bases ##
    MhdMultifluidBasesLoader,
    ## multifluid_calculator ##
    MhdMultifluidCalculator,
    ## multifluid_densities ##
    MhdMultifluidDensityLoader,
    ## multifluid_ionization ##
    MhdMultifluidIonizationLoader, saha_n1n0,
    ## multifluid_species ##
    Specie, SpecieList,
)



"""
File Purpose: vector derivatives (e.g. curl, div, grad)
"""
from ..quantity_loader import QuantityLoader
from ...dimensions import YZ_FROM_X
from ...tools import xarray_differentiate, xarray_isel

class VectorDerivativeLoader(QuantityLoader):
    '''vector derivatives (e.g. curl, div, grad).
    E.g. div_u --> div(u), i.e. du_x/dx + du_y/dy + du_z/dz.
    '''
    # div_{var}; optional "__{axes}"
    @known_pattern(r'div_(.*?[^_]+?)(__[xyz]{1,3})?', deps=[0], ignores_dims=['component'])
    def get_div(self, var, *, _match=None):
        '''divergence. 'div_{var}' --> div(var), i.e. du_x/dx + du_y/dy + du_z/dz.
        if component(s) is provided, only include that component(s) during the calculation.
            e.g. div_u__xy --> du_x/dx + du_y/dy.

        if self.slices nonempty, self.deriv_before_slice controls whether slicing occurs before or after div.
        '''
        var, axes = _match.groups()
        axes = 'xyz' if (axes is None) else axes[len('__'):]  # remove '__'. e.g. '__x' --> 'x'
        pre_slices, post_slices = self._pre_and_post_deriv_slices(list(axes))
        with self.using(slices=pre_slices):
            value = self(f'{var}_{axes}')   # [EFF] get all components at once, instead of one at a time.
            # [TODO][EFF] is it better to get one at a time, if using self.deriv_before_slice != False?
        comps = self.take_components(value, drop_labels=True)  # e.g. [var_x, var_y, var_z]
        diffs = [xarray_differentiate(val, str(x)) for val, x in zip(comps, axes)]
        if post_slices:
            diffs = [xarray_isel(diff, **post_slices) for diff in diffs]
        return sum(diffs)

    # grad_{var}
    @known_pattern(r'grad_(.+)', deps=[0])
    def get_grad(self, var, *, _match=None):
        '''gradient. 'grad_{var}' --> grad(var), i.e. (dn/dx, dn/dy, dn/dz).
        returned components are determined by self.component.
            (see also: the get_xyz pattern. E.g., 'grad_n_x' --> x component of grad(n).
            to get grad of a vector's component instead, use parentheses, e.g. 'grad_(u_x)')

        if self.slices nonempty, self.deriv_before_slice controls whether slicing occurs before or after div.
        '''
        var, = _match.groups()
        pre_slices, post_slices = self._pre_and_post_deriv_slices([str(c) for c in self.component_list()])
        with self.using(slices=pre_slices):
            value = self(var)   # for grad, var is a scalar.
        result = []
        for x in self.iter_component():
            diff = xarray_differentiate(value, str(x))
            result.append(self.assign_component_coord(diff, x))
        if post_slices:
            result = [xarray_isel(diff, **post_slices) for diff in result]
        return self.join_components(result)

    def curl_component(self, v, x, *, yz=None):
        '''return x component of curl(v).

        v: xarray.DataArray
            vector to take curl of.
            spatial derivatives will apply to dimensions ('x', 'y', 'z') (if they exist, else give 0).
            must include 'components' dimension including coordinates y and z.
        x: int, str, or Component
            tells component to get. if int or str, use self.components to get corresponding Component
        yz: None or iterable of two (int, str, or Component) objects
            the other two components; (x,y,z) should form a right-handed coordinate system.
            if not provided, infer from x.
        '''
        if yz is None: yz = YZ_FROM_X[x]
        y, z = yz
        vy, vz = self.take_components(v, yz)
        ddy_vz = xarray_differentiate(vz, str(y))
        ddz_vy = xarray_differentiate(vy, str(z))
        return self.assign_component_coord(ddy_vz - ddz_vy, x)

    # curl_{var}
    @known_pattern(r'curl_(.+)', deps=[0])
    def get_curl(self, var, *, _match=None):
        '''curl. 'curl_{var}' --> curl(var), i.e. (du_z/dy - du_y/dz, du_x/dz - du_z/dx, du_y/dx - du_x/dy).
        returned components are determined by self.component.
            (see also: the get_xyz pattern. E.g., curl_u_x --> x component of curl(u))

        if self.slices nonempty, self.deriv_before_slice controls whether slicing occurs before or after div.
        '''
        var, = _match.groups()
        if not self.component_is_iterable():  # single component of result,
            # so we only need some of the components of var
            x = self.component
            y, z = yz = YZ_FROM_X[x]
            pre_slices, post_slices = self._pre_and_post_deriv_slices([y, z])
            with self.using(slices=pre_slices):
                v = self(f'{var}_{y}{z}')
            result = self.curl_component(v, x, yz=yz)
            if post_slices:
                result = xarray_isel(result, **post_slices)
            return result
        # else: multiple components of result -- we do need all the components of var.
        pre_slices, post_slices = self._pre_and_post_deriv_slices(['x', 'y', 'z'])
        with self.using(slices=pre_slices):
            v = self(f'{var}_xyz')
        result = [self.curl_component(v, x) for x in self.component]
        if post_slices:
            result = [xarray_isel(diff, **post_slices) for diff in result]
        return self.join_components(result)
